/*
 *
 *                         PCI I/O Board Device Driver
 *
 *                               For Kernel 2.4.x
 *
 *
 *	Driver for PCI device driver
 *	Based on drivers/char/serial.c, by Linus Torvalds, Theodore Ts'o.
 *
 * 	This program is free software; you can redistribute it and/or modify
 * 	it under the terms of the GNU General Public License as published by
 * 	the Free Software Foundation; either version 2 of the License, or
 * 	(at your option) any later version.
 *
 * 	This program is distributed in the hope that it will be useful,
 * 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 * 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * 	GNU General Public License for more details.
 *
 * 	You should have received a copy of the GNU General Public License
 * 	along with this program; if not, write to the Free Software
 * 	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 *                                                    		Version: 1.05
 *                                                        Date: 2006/12/01
 */
#ifdef 		MODVERSIONS
#ifndef 	MODULE
#define 	MODULE
#endif
#endif

#ifdef MODULE
#include <linux/config.h>
#ifdef MODVERSIONS
#include <linux/modversions.h>
#endif
#include <linux/module.h>
#else
#define	MOD_INC_USE_COUNT
#define MOD_DEC_USE_COUNT
#endif



#include <linux/version.h>
#include <linux/autoconf.h>
#include <linux/errno.h>
#include <linux/signal.h>
#include <linux/sched.h>
#include <linux/timer.h>
#include <linux/interrupt.h>
#include <linux/tty.h>
#include <linux/tty_flip.h>
#include <linux/serial.h>
#include <linux/serial_reg.h>
#include <linux/major.h>
#include <linux/string.h>
#include <linux/fcntl.h>
#include <linux/ptrace.h>
#include <linux/ioport.h>
#include <linux/mm.h>
#include <linux/smp_lock.h>
#include <linux/delay.h>


#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/wait.h>
#include <linux/tty_driver.h>
#include <linux/signal.h>
#include <linux/pci.h>
//#include <linux/workqueue.h>
#include <linux/circ_buf.h>



#include <asm/uaccess.h>
#include <asm/system.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/segment.h>
#include <asm/serial.h>
#include <asm/bitops.h>

#include "linux24.h"


#ifdef CONFIG_PCI
#include <linux/pci.h>
#ifndef PCI_ANY_ID
#define PCI_ANY_ID (~0)
#endif
#endif



MODULE_AUTHOR(DRIVER_AUTHOR);
MODULE_DESCRIPTION(DRIVER_DESC);
MODULE_LICENSE("GPL");




enum	{
// 1S
 	LINUX_BOARD_4027 = 1,

// 2S
	LINUX_BOARD_4036,
	LINUX_BOARD_4037,
	LINUX_BOARD_8139,

// 4S
	LINUX_BOARD_4055,
	LINUX_BOARD_4056,
	LINUX_BOARD_8159,

// 8S
	LINUX_BOARD_4066,
	LINUX_BOARD_8169,
};

static unsigned int linux_board_name[] =
{
// 1S
 	4027,

// 2S
	4036,
	4037,
	8139,

// 4S
	4055,
	4056,
	8159,

// 8S
	4066,
	8169,
};

static int linux_numports[] =
{
// 1S
 	1,	// 4027

// 2S
	2,	// 4036
	2,	// 4037
	2,	// 8139

// 4S
	4,	// 4055
	4,	// 4056
	4,	// 8159

// 8S
	8,	// 4066
	8,	// 8169
};


typedef struct
{
	unsigned short vendor;
	unsigned short device;
	unsigned short subvendor;
	unsigned short subdevice;
	unsigned short driver_data;
} linux_pciInfo;


static linux_pciInfo linux_pcibrds[] =
{

// 1S
	{VENDOR_ID_LINUX24, DEVICE_ID_LINUX24 ,SUBVENDOR_ID_LINUX24, SUBDEVICE_ID_LINUX24_4027A,  LINUX_BOARD_4027},
	{VENDOR_ID_LINUX24, DEVICE_ID_LINUX24 ,SUBVENDOR_ID_LINUX24, SUBDEVICE_ID_LINUX24_4027D,  LINUX_BOARD_4027},

// 2S
	{VENDOR_ID_LINUX24, DEVICE_ID_LINUX24 ,SUBVENDOR_ID_LINUX24, SUBDEVICE_ID_LINUX24_4036A,  LINUX_BOARD_4036},
	{VENDOR_ID_LINUX24, DEVICE_ID_LINUX24 ,SUBVENDOR_ID_LINUX24, SUBDEVICE_ID_LINUX24_4036D,  LINUX_BOARD_4036},
	{VENDOR_ID_LINUX24, DEVICE_ID_LINUX24 ,SUBVENDOR_ID_LINUX24, SUBDEVICE_ID_LINUX24_4037A,  LINUX_BOARD_4037},
	{VENDOR_ID_LINUX24, DEVICE_ID_LINUX24 ,SUBVENDOR_ID_LINUX24, SUBDEVICE_ID_LINUX24_4037D,  LINUX_BOARD_4037},
	{VENDOR_ID_LINUX24, DEVICE_ID_LINUX24 ,SUBVENDOR_ID_LINUX24, SUBDEVICE_ID_LINUX24_8139,   LINUX_BOARD_8139},

// 4S
	{VENDOR_ID_LINUX24, DEVICE_ID_LINUX24 ,SUBVENDOR_ID_LINUX24, SUBDEVICE_ID_LINUX24_4055WN, LINUX_BOARD_4055},
	{VENDOR_ID_LINUX24, DEVICE_ID_LINUX24 ,SUBVENDOR_ID_LINUX24, SUBDEVICE_ID_LINUX24_4056A,  LINUX_BOARD_4056},
	{VENDOR_ID_LINUX24, DEVICE_ID_LINUX24 ,SUBVENDOR_ID_LINUX24, SUBDEVICE_ID_LINUX24_4056D,  LINUX_BOARD_4056},
	{VENDOR_ID_LINUX24, DEVICE_ID_LINUX24 ,SUBVENDOR_ID_LINUX24, SUBDEVICE_ID_LINUX24_8159,   LINUX_BOARD_8159},

// 8S
 	{VENDOR_ID_LINUX24, DEVICE_ID_LINUX24 ,SUBVENDOR_ID_LINUX24, SUBDEVICE_ID_LINUX24_4066A,  LINUX_BOARD_4066},
 	{VENDOR_ID_LINUX24, DEVICE_ID_LINUX24 ,SUBVENDOR_ID_LINUX24, SUBDEVICE_ID_LINUX24_4066R,  LINUX_BOARD_4066},
	{VENDOR_ID_LINUX24, DEVICE_ID_LINUX24 ,SUBVENDOR_ID_LINUX24, SUBDEVICE_ID_LINUX24_8169,   LINUX_BOARD_8169},
	{0}
};



// PCI Board struct
struct linux_board_hwconf
{
	unsigned int		    board_name;
	unsigned int		    bus_number;
	unsigned int		    dev_number;
	int							    board_type;
	int							    ports;
	int							    irq;
	int							    vector;
	int							    vector_mask;
	int							    uart_type;
	int							    ioaddr[LINUX24_MAX_PORTS_PER_BOARD];
	struct 	pci_dev	    *pdev;
	int							    ChipFlag[LINUX24_MAX_PORTS_PER_BOARD];
	int							    baud_base[LINUX24_MAX_PORTS_PER_BOARD];

	unsigned int		    remap;	//add by CHL 2006/10/02
	unsigned short int 	subdeviceID; //add by CHL 2006/10/02    
};

// port struct
struct linux_port_hwconf
{
	unsigned int		  board_name;
	int							  board_type;
	int							  baud_rate;
	unsigned char		  FC_flag;
	unsigned int		  bus_number;
	unsigned int		  dev_number;
	int							  port;
	int							  base;
	int							  irq;
	int							  vector;
	int							  vector_mask;
	int							  rx_trigger;
	int							  baud_base;
	int							  flags;
	int							  type;
	struct tty_struct	*tty;
	int							  read_status_mask;
	int							  ignore_status_mask;
	int							  xmit_fifo_size;
	int							  custom_divisor;
	unsigned char		  x_char;
	int							  close_delay;
	unsigned short	  closing_wait;
	unsigned char		  IER;
	unsigned char		  MCR;
	unsigned char 	  FCR;
	unsigned long		  event;
	int							  count;
	int							  blocked_open;
	long						  session;
	long						  pgrp;
	unsigned char		  *tx_buf;
	int							  tx_buf_cont;
	int							  tx_buf_head;
	int							  tx_buf_tail;

	struct tq_struct 		tqueue;

	struct termios			normal_termios;
	struct termios			callout_termios;

	wait_queue_head_t 	open_wait;
	wait_queue_head_t 	close_wait;
	wait_queue_head_t 	delta_msr_wait;
	struct async_icount	icount;

 	int							  timeout;
	int							  ChipFlag;
 	unsigned char		  ldisc_stop_rx;
	unsigned char     err_shadow;
	spinlock_t			  slock;

	unsigned int		    remap;	//add by CHL 2006/10/02
	unsigned short int 	subdeviceID; //add by CHL 2006/10/02   
};


static struct tty_driver		 linux_tty_sdriver;
static struct tty_driver		 linux_tty_cdriver;
static int		linux_refcount;



static struct tty_struct		*linux_tty[LINUX24_MAX_PORTS_TOTAL];
static struct termios				*linux_termios[LINUX24_MAX_PORTS_TOTAL];
static struct termios				*linux_termios_locked[LINUX24_MAX_PORTS_TOTAL];

static struct linux_board_hwconf	 linux_pci_board[LINUX24_MAX_BOARDS];
static struct linux_port_hwconf		 linux_port_table[LINUX24_MAX_PORTS_TOTAL];

static unsigned char linux_msr[LINUX24_MAX_PORTS_TOTAL];





static int  baudrate_setup(struct linux_port_hwconf *, unsigned int);
static int  termios_setup(struct linux_port_hwconf *, struct termios *);
static int  linux_startup(struct linux_port_hwconf *);
static int  linux_block_til_ready(struct tty_struct *, struct file *, struct linux_port_hwconf *);
static void linux_shutdown(struct linux_port_hwconf *);
static void linux_wait_until_sent(struct tty_struct *, int);
static void linux_flush_chars(struct tty_struct *);
static int  linux_open(struct tty_struct *, struct file *);
static void linux_close(struct tty_struct *, struct file *);
static int  linux_write(struct tty_struct *, int, const unsigned char *, int);
static void linux_put_char(struct tty_struct *, unsigned char);
static int  linux_write_room(struct tty_struct *);
static int  linux_chars_in_buffer(struct tty_struct *);
static void linux_flush_buffer(struct tty_struct *);
static void linux_throttle(struct tty_struct *);
static void linux_unthrottle(struct tty_struct *);
static void linux_stop(struct tty_struct *);
static void linux_start(struct tty_struct *);
static void linux_rs_break(struct tty_struct *, int);
static void linux_interrupt(int, void *, struct pt_regs *);
static void linux_send_break(struct linux_port_hwconf *, int);
static int  linux_get_modem_info(struct linux_port_hwconf *, unsigned int *);
static int  linux_set_modem_info(struct linux_port_hwconf *, unsigned int, unsigned int *);
static int  linux_get_uart_info(struct linux_port_hwconf *, struct serial_struct *);
static int  linux_set_uart_info(struct linux_port_hwconf *, struct serial_struct *);
static int  linux_get_lsr_info(struct linux_port_hwconf *, unsigned int *);
static void linux_set_termios(struct tty_struct *, struct termios *);
static int  linux_ioctl(struct tty_struct *, struct file *, unsigned int, unsigned long);
static void linux_do_softint(void *);
static int  linux_get_PCI_conf(int, struct linux_board_hwconf *);
static int  linux_board_init(int, struct linux_board_hwconf *);
void 				linux_hangup(struct tty_struct *);

static inline void linux_modem_status_intr(struct linux_port_hwconf *, unsigned char);
static inline void linux_data_transmit(struct linux_port_hwconf *);
static inline void linux_data_receive(struct linux_port_hwconf *, int *, unsigned char);





static int baudrate_setup(struct linux_port_hwconf *info, unsigned int new_speed)
{
	int	dlldlm = 0, retval = 0;
	unsigned char	lcr;
	unsigned long	flag;


	if (!info->tty || !info->tty->termios)
  {
		return retval;
  }

	if (!(info->base))
  {
		return retval;
  }

 	info->baud_rate = new_speed;

	if (new_speed)
	{
	  dlldlm = (info->baud_base * 16) / (new_speed * 16);

	  if (dlldlm == 0)
    {
	  	dlldlm = 1;
    }
	}
	else
  {
   	dlldlm = 0;
  }


	info->timeout = ((info->xmit_fifo_size * HZ * 10 * dlldlm) / info->baud_base);

	if (dlldlm)
	{
		spin_lock_irqsave(&info->slock, flag);
 		info->MCR |= UART_MCR_DTR;
 		outb(info->MCR, info->base + UART_MCR);
		spin_unlock_irqrestore(&info->slock,flag);
	}
	else
	{
		spin_lock_irqsave(&info->slock, flag);
 		info->MCR &= ~UART_MCR_DTR;
 		outb(info->MCR, info->base + UART_MCR);
		spin_unlock_irqrestore(&info->slock,flag);
 		return retval;
	}

 	lcr = inb(info->base + UART_LCR);

	outb(lcr | UART_LCR_DLAB, info->base + UART_LCR);
	outb(dlldlm & 0xff, info->base + UART_DLL);
	outb(dlldlm >> 8  , info->base + UART_DLM);
	outb(lcr, info->base + UART_LCR);
	return retval;
}


static int termios_setup(struct linux_port_hwconf *info, struct termios *old_termios)
{
	unsigned int c_cflag, baudrate = 0;
	unsigned char	lcr = 0;
  int retval = 0;

	if (!info->tty || !info->tty->termios)
  {
		return retval;
  }

	c_cflag = info->tty->termios->c_cflag;

	if (!(info->base))
  {
		return retval;
  }

	switch (c_cflag & (CBAUD | CBAUDEX))
	{
   	case B921600	: baudrate = 921600;	break;
   	case B460800	: baudrate = 460800;	break;
   	case B230400	: baudrate = 230400;	break;
   	case B115200	: baudrate = 115200;	break;
   	case B57600		: baudrate = 57600;		break;
   	case B38400		: baudrate = 38400;		break;
   	case B19200		: baudrate = 19200;		break;
   	case B9600		: baudrate = 9600;		break;
   	case B4800		: baudrate = 4800;		break;
   	case B2400		: baudrate = 2400;		break;
   	case B1200		: baudrate = 1200;		break;
   	case B300			: baudrate = 300;			break;
   	case B110			: baudrate = 110;			break;
   	default				: baudrate = 0;				break;
	}

  baudrate_setup(info, baudrate);

	switch (c_cflag & CSIZE)
	{
		case CS5: 	lcr = 0x00; break;
		case CS6: 	lcr = 0x01; break;
		case CS7: 	lcr = 0x02; break;
		case CS8: 	lcr = 0x03; break;
		default:  	lcr = 0x03; break;
	}

	if (c_cflag & CSTOPB)
	{
		lcr |= 0x04;
	}

	if (c_cflag & PARENB)
	{
		if (c_cflag & PARODD)
    {
			lcr |= 0x08;
    }
		else
    {
			lcr |= 0x18;
    }
	}

	if (c_cflag & CMSPAR)
	{
		lcr |= UART_LCR_SPAR;
	}

	info->FCR = UART_FCR_ENABLE_FIFO;

	if (info->ChipFlag == LINUX24_1889_HWID)
	{
		switch (info->rx_trigger)
		{
			case 8: 	info->FCR |= 0x21;	break;
			case 16: 	info->FCR |= 0x61;	break;
			case 24: 	info->FCR |= 0xa1;	break;
			case 28: 	info->FCR |= 0xe1;	break;
			default: 	info->FCR |= 0x61; 	break;
		}
	}

	if (info->ChipFlag  == LINUX24_1699_HWID)
	{
		switch (info->rx_trigger)
		{
			case 16: 	info->FCR |= 0x31; 	break;
			case 32: 	info->FCR |= 0x71; 	break;
			case 48: 	info->FCR |= 0xb1; 	break;
			case 56: 	info->FCR |= 0xf1; 	break;
			default:  info->FCR |= 0x31;	break;
		}
	}

	info->IER &= ~UART_IER_MSI;
	info->MCR &= ~UART_MCR_AFE;

	if (c_cflag & CRTSCTS)
	{
		info->FC_flag |= LINUX24_FLOWCTL_HW_AUTO;
		info->IER     |= UART_IER_MSI;
 		info->MCR     |= UART_MCR_AFE;
	}
	else
	{
		info->FC_flag &= ~LINUX24_FLOWCTL_HW_AUTO;
	}

	if (c_cflag & CLOCAL)
	{
		info->flags &= ~ASYNC_CHECK_CD;
	}
	else
	{
		info->flags |= ASYNC_CHECK_CD;
		info->IER   |= UART_IER_MSI;
	}

	if (I_IXON(info->tty) || I_IXOFF(info->tty))
	{
		info->FC_flag |= LINUX24_FLOWCTL_SW;
	}
	else
	{
		info->FC_flag &= ~LINUX24_FLOWCTL_SW;
	}


	info->read_status_mask = UART_LSR_DR | UART_LSR_OE | UART_LSR_THRE;

	if (I_INPCK(info->tty))
	{
		info->read_status_mask |= UART_LSR_FE | UART_LSR_PE;
	}

	if (I_BRKINT(info->tty) || I_PARMRK(info->tty))
	{
	  info->read_status_mask |= UART_LSR_BI;
	}

	info->ignore_status_mask = 0;

	if (I_IGNBRK(info->tty))
	{
	  info->ignore_status_mask |= UART_LSR_BI;
	  info->read_status_mask   |= UART_LSR_BI;

    if (I_IGNPAR(info->tty))
		{
			info->ignore_status_mask |= UART_LSR_OE | UART_LSR_PE | UART_LSR_FE;
			info->read_status_mask   |= UART_LSR_OE | UART_LSR_PE | UART_LSR_FE;
   	}
	}

	outb(info->MCR, info->base + UART_MCR);
	outb(info->IER, info->base + UART_IER);

	outb(info->FCR, info->base + UART_FCR);
	outb(lcr, info->base + UART_LCR);

  return retval;
}


static int linux_startup(struct linux_port_hwconf *info)
{
	unsigned long	page, flag;

	page = __get_free_page(GFP_KERNEL);

	if (!page)
  {
		return(-ENOMEM);
  }


	spin_lock_irqsave(&info->slock, flag);


	if (info->flags & ASYNC_INITIALIZED)
	{
	 	free_page(page);
		spin_unlock_irqrestore(&info->slock,flag);
   	return(0);
	}

	if (!info->base || !info->type)
	{
	 	if (info->tty)
    {
			set_bit(TTY_IO_ERROR, &info->tty->flags);
    }

   	free_page(page);
		spin_unlock_irqrestore(&info->slock,flag);
   	return(0);
	}

	if (info->tx_buf)
  {
	  free_page(page);
  }
	else
  {
	  info->tx_buf = (unsigned char *)page;
  }


 	outb((UART_FCR_CLEAR_RCVR | UART_FCR_CLEAR_XMIT), info->base + UART_FCR);

	if (inb(info->base + UART_LSR) == 0xff)
	{
		spin_unlock_irqrestore(&info->slock,flag);
    if (capable(CAP_SYS_ADMIN))
		{
			if (info->tty)
      {
		    set_bit(TTY_IO_ERROR, &info->tty->flags);
      }
			return(0);
	  }
		else
    {
			return(-ENODEV);
    }
	}

 	(void)inb(info->base + UART_LSR);
 	(void)inb(info->base + UART_RX );
 	(void)inb(info->base + UART_IIR);
 	(void)inb(info->base + UART_MSR);

 	outb(UART_LCR_WLEN8, info->base + UART_LCR);

 	info->MCR = UART_MCR_DTR | UART_MCR_RTS;
	outb(info->MCR, info->base + UART_MCR);

  /* mark by CHL 2006/11/23
	info->IER = UART_IER_MSI | UART_IER_RLSI | UART_IER_RDI;

	if (info->ChipFlag  == LINUX24_1699_HWID)
	{
  	info->IER |= LINUX24_1699_CLK_DIVIDER_DISABLE;
 	}

 	outb(info->IER, info->base + UART_IER);

 	(void)inb(info->base + UART_LSR);
 	(void)inb(info->base + UART_RX );
 	(void)inb(info->base + UART_IIR);
 	(void)inb(info->base + UART_MSR);
  */

  
  if (info->tty)
  {
 		clear_bit(TTY_IO_ERROR, &info->tty->flags);
  }

  info->tx_buf_cont = info->tx_buf_head = info->tx_buf_tail = 0;

	spin_unlock_irqrestore(&info->slock,flag);

 	termios_setup(info, 0);

  // move to here, setting IER after set baudrate...
  info->IER = UART_IER_MSI | UART_IER_RLSI | UART_IER_RDI;

	if (info->ChipFlag  == LINUX24_1699_HWID)
	{
  	info->IER |= LINUX24_1699_CLK_DIVIDER_DISABLE;
 	}

 	outb(info->IER, info->base + UART_IER);

 	(void)inb(info->base + UART_LSR);
 	(void)inb(info->base + UART_RX );
 	(void)inb(info->base + UART_IIR);
 	(void)inb(info->base + UART_MSR);

  
	info->flags |= ASYNC_INITIALIZED;
	return 0;
}


static int linux_block_til_ready(struct tty_struct *tty, struct file * filp, struct linux_port_hwconf *info)
{
	DECLARE_WAITQUEUE(wait, current);
	int	retval, do_clocal = 0;
	unsigned long flag;

	if ((filp->f_flags & O_NONBLOCK) || (tty->flags & (1 <<TTY_IO_ERROR)))
	{
    
	  if (info->flags & ASYNC_CALLOUT_ACTIVE)
    {
			return(-EBUSY);
    }

   	info->flags |= ASYNC_NORMAL_ACTIVE;
   	return(0);
	}


	if (info->flags & ASYNC_CALLOUT_ACTIVE)
	{
	  if (info->normal_termios.c_cflag & CLOCAL)
    {
		  do_clocal = 1;
    }
	}
	else
	{
	  if (tty->termios->c_cflag & CLOCAL)
    {
			do_clocal = 1;
    }
	}


	retval = 0;
	add_wait_queue(&info->open_wait, &wait);

	spin_lock_irqsave(&info->slock, flag);

	if (!tty_hung_up_p(filp))
  {
		info->count--;
  }

	spin_unlock_irqrestore(&info->slock,flag);

	info->blocked_open++;

	while (1)
	{
		spin_lock_irqsave(&info->slock, flag);

	  if (!(info->flags & ASYNC_CALLOUT_ACTIVE))
    {
			outb(inb(info->base + UART_MCR) | UART_MCR_DTR | UART_MCR_RTS, info->base + UART_MCR);
    }

		spin_unlock_irqrestore(&info->slock,flag);

  	set_current_state(TASK_INTERRUPTIBLE);

  	if (tty_hung_up_p(filp) || !(info->flags & ASYNC_INITIALIZED))
		{
			if (info->flags & ASYNC_HUP_NOTIFY)
      {
	   		retval = -EAGAIN;
      }
			else
      {
	    	retval = -ERESTARTSYS;
      }

			break;
   	}
 
	  if (!(info->flags & ASYNC_CALLOUT_ACTIVE) &&
		 		!(info->flags & ASYNC_CLOSING) &&
		 		(do_clocal || (inb(info->base + UART_MSR) & UART_MSR_DCD)))
    {
			break;
    }
    
	  if (signal_pending(current))
		{
			retval = -ERESTARTSYS;
			break;
	  }

	  schedule();
	}

	set_current_state(TASK_RUNNING);
	remove_wait_queue(&info->open_wait, &wait);

 	if (!tty_hung_up_p(filp))
  {
 		info->count++;
  }

	info->blocked_open--;

	if (retval)
  {
		return(retval);
  }

	info->flags |= ASYNC_NORMAL_ACTIVE;
	return(0);
}


static void linux_shutdown(struct linux_port_hwconf *info)
{
	unsigned long flag;

	if (!(info->flags & ASYNC_INITIALIZED))
  {
		return;
  }

	spin_lock_irqsave(&info->slock, flag);

	wake_up_interruptible(&info->delta_msr_wait);

	if (info->tx_buf)
	{
		free_page((unsigned long)info->tx_buf);
		info->tx_buf = NULL;
	}

	info->IER = 0;
	outb(0x00 , info->base + UART_IER);

	if (!info->tty || (info->tty->termios->c_cflag & HUPCL))
  {
		info->MCR &= ~(UART_MCR_DTR | UART_MCR_RTS);
  }

	outb(info->MCR, info->base + UART_MCR);

	outb((UART_FCR_CLEAR_RCVR|UART_FCR_CLEAR_XMIT), info->base + UART_FCR);

	(void)inb(info->base + UART_RX);

	if (info->tty)
  {
		set_bit(TTY_IO_ERROR, &info->tty->flags);
  }

	info->flags &= ~ASYNC_INITIALIZED;

	spin_unlock_irqrestore(&info->slock,flag);
}


static void linux_wait_until_sent(struct tty_struct *tty, int timeout)
{
	struct linux_port_hwconf *info = (struct linux_port_hwconf *)tty->driver_data;

	unsigned long orig_jiffies, char_time;
	int lsr;

	if (info->type == PORT_UNKNOWN)
  {
		return;
  }

	if (info->xmit_fifo_size == 0)
  {
		return;
  }

	orig_jiffies = jiffies;
	char_time = (info->timeout - HZ/50) / info->xmit_fifo_size;
	char_time = char_time / 5;

	if (char_time == 0)
  {
		char_time = 1;
  }

	if (timeout && timeout < char_time)
  {
		char_time = timeout;
  }

	if (!timeout || timeout > 2*info->timeout)
  {
		timeout = 2*info->timeout;
  }

	while (!((lsr = inb(info->base+ UART_LSR)) & UART_LSR_TEMT))
	{
		set_current_state(TASK_INTERRUPTIBLE);
		schedule_timeout(char_time);
    
		if (signal_pending(current))
    {
			break;
    }

		if (timeout && time_after(jiffies, orig_jiffies + timeout))
    {
			break;
    }
	}

	set_current_state(TASK_RUNNING);
}


static void linux_flush_chars(struct tty_struct *tty)
{
	struct linux_port_hwconf *info = (struct linux_port_hwconf *)tty->driver_data;

	unsigned long	flag;

	if (info->tx_buf_cont<=0||tty->stopped||!info->tx_buf||(tty->hw_stopped))
  {
		return;
  }

	spin_lock_irqsave(&info->slock, flag);

	info->IER |= UART_IER_THRI;
	outb(info->IER, info->base + UART_IER);

	spin_unlock_irqrestore(&info->slock,flag);
}


static int linux_open(struct tty_struct *tty, struct file *filp)
{
	struct linux_port_hwconf *info;
	int retval, line;

 	line = MINOR(tty->device) - tty->driver.minor_start;

	if ((line < 0) || (line >= LINUX24_MAX_PORTS_TOTAL))
  {
		return(-ENODEV);
  }

	info = linux_port_table + line;

	if (!info->base)
  {
   	return(-ENODEV);
  }

 	tty->driver_data = info;
 	info->tty = tty;

 	retval = linux_startup(info);

 	if (retval)
  {
 		return(retval);
  }

 	retval = linux_block_til_ready(tty, filp, info);

 	if (retval)
  {
 		return(retval);
  }

	info->count++;

	MOD_INC_USE_COUNT;

	if ((info->count == 1) && (info->flags & ASYNC_SPLIT_TERMIOS))
	{
   	if (tty->driver.subtype  == SERIAL_TYPE_NORMAL)
    {
			*tty->termios = info->normal_termios;
    }
   	else
    {
			*tty->termios = info->callout_termios;
    }

		termios_setup(info, 0);
	}

	info->session   = current->session;

 	info->pgrp	= current->pgrp;
	clear_bit(TTY_DONT_FLIP, &tty->flags);
 	tty->low_latency = 1;

	return (0);
}


static void linux_close(struct tty_struct *tty, struct file *filp)
{
	struct linux_port_hwconf *info = (struct linux_port_hwconf *)tty->driver_data;
 	unsigned long	timeout, flag;
	int	line;

 	line = MINOR(tty->device) - tty->driver.minor_start;

	if ((line < 0) || (line >=LINUX24_MAX_PORTS_TOTAL))
  {
		return;
  }

	if (!info)
  {
   	return;
  }

	spin_lock_irqsave(&info->slock, flag);

	if (tty_hung_up_p(filp))
	{
		spin_unlock_irqrestore(&info->slock,flag);
 		MOD_DEC_USE_COUNT;
		return;
	}

#if (LINUX_VERSION_CODE <= KERNEL_VERSION(2,4,21))
	if ((tty->count == 1) && (info->count != 1))
	{
#else
	if ((atomic_read(&tty->count) == 1) && (info->count != 1))
	{
#endif
   	info->count = 1;
	}

	if (--info->count < 0)
  {
   	info->count = 0;
  }

	if (info->count)
	{
		spin_unlock_irqrestore(&info->slock,flag);
   	MOD_DEC_USE_COUNT;
   	return;
	}

	info->flags |= ASYNC_CLOSING;
	spin_unlock_irqrestore(&info->slock,flag);

	if (info->flags & ASYNC_NORMAL_ACTIVE)
  {
	  info->normal_termios = *tty->termios;
  }

	if (info->flags & ASYNC_CALLOUT_ACTIVE)
  {
		info->callout_termios = *tty->termios;
  }

	tty->closing = 1;

	if (info->closing_wait != ASYNC_CLOSING_WAIT_NONE)
  {
		tty_wait_until_sent(tty, info->closing_wait);
  }

	info->IER &= ~UART_IER_RLSI;

	if (info->flags & ASYNC_INITIALIZED)
	{
		outb(info->IER, info->base + UART_IER);
	  timeout = jiffies + HZ;

	  while (!(inb(info->base + UART_LSR) & UART_LSR_TEMT))
		{
			set_current_state(TASK_INTERRUPTIBLE);
			schedule_timeout(5);

			if (time_after(jiffies, timeout))
      {
		 		break;
      }
 		}
	}

	linux_shutdown(info);

	if (tty->driver.flush_buffer)
  {
		tty->driver.flush_buffer(tty);
  }

	if (tty->ldisc.flush_buffer)
  {
		tty->ldisc.flush_buffer(tty);
  }



	tty->closing 	= 0;
	info->event 	= 0;
	info->tty 	= NULL;

	if (info->blocked_open)
	{
	  if (info->close_delay)
		{
			set_current_state(TASK_INTERRUPTIBLE);
			schedule_timeout(info->close_delay);
	  }
	  wake_up_interruptible(&info->open_wait);
	}

	info->flags &= ~(ASYNC_NORMAL_ACTIVE | ASYNC_CALLOUT_ACTIVE | ASYNC_CLOSING);

	wake_up_interruptible(&info->close_wait);
	MOD_DEC_USE_COUNT;
}


static int linux_write(struct tty_struct *tty, int from_user, const unsigned char *buf, int count)
{
	struct linux_port_hwconf *info = (struct linux_port_hwconf *)tty->driver_data;
	int	c, total = 0;
	unsigned long	flag;

	if (!tty || !info->tx_buf)
  {
 	  return(0);
  }

  
	while (1)
	{
 		c = MIN(count, MIN(SERIAL_XMIT_SIZE - info->tx_buf_cont - 1, SERIAL_XMIT_SIZE - info->tx_buf_head));
		if (c <= 0)
    {
			break;
    }

 		if (from_user)
		{
			if (copy_from_user( info->tx_buf + info->tx_buf_head , buf, c) == c)
			{
				total = -EFAULT;
	 			break;
			}
 		}
		else
		{
			memcpy(info->tx_buf + info->tx_buf_head, buf, c);
		}

		spin_lock_irqsave(&info->slock, flag);
 		info->tx_buf_head   = (info->tx_buf_head + c) & (SERIAL_XMIT_SIZE - 1);
 		info->tx_buf_cont  += c;
		spin_unlock_irqrestore(&info->slock,flag);

 		buf   += c;
 		count -= c;
 		total += c;
	}

 	if (info->tx_buf_cont && !tty->stopped && !(info->IER & UART_IER_THRI))
	{
		spin_lock_irqsave(&info->slock, flag);
   	info->IER |= UART_IER_THRI;
   	outb(info->IER, info->base + UART_IER);
		spin_unlock_irqrestore(&info->slock,flag);
	}

	return total;
}


static void linux_put_char(struct tty_struct *tty, unsigned char ch)
{
	struct linux_port_hwconf *info = (struct linux_port_hwconf *)tty->driver_data;
	unsigned long	flag;

	if (!tty || !info->tx_buf)
  {
	 	return;
  }

	if (info->tx_buf_cont >= SERIAL_XMIT_SIZE - 1)
  {
	 	return;
  }

	spin_lock_irqsave(&info->slock, flag);
	info->tx_buf[info->tx_buf_head++] = ch;
	info->tx_buf_head &= SERIAL_XMIT_SIZE - 1;
	info->tx_buf_cont++;
	spin_unlock_irqrestore(&info->slock,flag);

	if (!tty->stopped && !(info->IER & UART_IER_THRI))
	{
		spin_lock_irqsave(&info->slock, flag);
   	info->IER |= UART_IER_THRI;
   	outb(info->IER, info->base + UART_IER);
		spin_unlock_irqrestore(&info->slock,flag);
	}
}


static int linux_write_room(struct tty_struct *tty)
{
	struct linux_port_hwconf *info = (struct linux_port_hwconf *)tty->driver_data;
	int	ret;

	ret = SERIAL_XMIT_SIZE - info->tx_buf_cont - 1;

	if (ret < 0)
  {
		ret = 0;
  }

	return(ret);
}


static int linux_chars_in_buffer(struct tty_struct *tty)
{
	struct linux_port_hwconf *info = (struct linux_port_hwconf *)tty->driver_data;

	return(info->tx_buf_cont);
}


static void linux_flush_buffer(struct tty_struct *tty)
{
	struct linux_port_hwconf *info = (struct linux_port_hwconf *)tty->driver_data;
	unsigned long	flag;

	spin_lock_irqsave(&info->slock, flag);
	info->tx_buf_cont = info->tx_buf_head = info->tx_buf_tail = 0;
	spin_unlock_irqrestore(&info->slock,flag);

	outb((info->FCR | UART_FCR_CLEAR_RCVR | UART_FCR_CLEAR_XMIT), info->base + UART_FCR);
	outb(info->FCR, info->base+UART_FCR);

	wake_up_interruptible(&tty->write_wait);

	if ((tty->flags & (1 << TTY_DO_WRITE_WAKEUP)) && tty->ldisc.write_wakeup)
  {
		(tty->ldisc.write_wakeup)(tty);
  }
}


static void linux_throttle(struct tty_struct *tty)
{
	struct linux_port_hwconf *info = (struct linux_port_hwconf *)tty->driver_data;

  info->ldisc_stop_rx = 1;

	if (info->FC_flag & LINUX24_FLOWCTL_SW)
	{
		info->x_char = STOP_CHAR(tty);
 		outb(info->x_char, info->base + UART_TX);
		info->icount.tx++;
		info->IER |= UART_IER_THRI;
		outb(info->IER, info->base + UART_IER);
	}

	if (info->FC_flag & LINUX24_FLOWCTL_HW_AUTO)
	{
 		info->MCR &= ~UART_MCR_AFE;
 		info->MCR &= ~UART_MCR_RTS;
		outb(info->MCR, info->base + UART_MCR);
	}
}


static void linux_unthrottle(struct tty_struct *tty)
{
	struct linux_port_hwconf *info = (struct linux_port_hwconf *)tty->driver_data;
  info->ldisc_stop_rx = 0;

  if (info->FC_flag & LINUX24_FLOWCTL_SW)
	{
		info->x_char = START_CHAR(tty);
 		outb(info->x_char, info->base + UART_TX);
		info->icount.tx++;
		info->IER |= UART_IER_THRI;
		outb(info->IER, info->base + UART_IER);
	}

	if (info->FC_flag & LINUX24_FLOWCTL_HW_AUTO)
	{
 		info->MCR |= UART_MCR_AFE;
 		info->MCR |= UART_MCR_RTS;
		outb(info->MCR, info->base + UART_MCR);
	}
}


static void linux_stop(struct tty_struct *tty)
{
	struct linux_port_hwconf *info = (struct linux_port_hwconf *)tty->driver_data;
	unsigned long	flag;

	spin_lock_irqsave(&info->slock, flag);

	if (info->IER & UART_IER_THRI)
	{
   	info->IER &= ~UART_IER_THRI;
   	outb(info->IER, info->base + UART_IER);
	}

	spin_unlock_irqrestore(&info->slock,flag);
}


static void linux_start(struct tty_struct *tty)
{
	struct linux_port_hwconf *info = (struct linux_port_hwconf *)tty->driver_data;
	unsigned long	flag;

	spin_lock_irqsave(&info->slock, flag);

	if (info->tx_buf_cont && info->tx_buf && !(info->IER & UART_IER_THRI))
	{
   	info->IER |= UART_IER_THRI;
   	outb(info->IER, info->base + UART_IER);
	}

	spin_unlock_irqrestore(&info->slock,flag);
}


void linux_hangup(struct tty_struct *tty)
{
	struct linux_port_hwconf *info = (struct linux_port_hwconf *)tty->driver_data;

	linux_flush_buffer(tty);
	linux_shutdown(info);
	info->event = 0;
	info->count = 0;
	info->flags &= ~(ASYNC_NORMAL_ACTIVE | ASYNC_CALLOUT_ACTIVE);
	info->tty = NULL;
	wake_up_interruptible(&info->open_wait);
}


static void linux_rs_break(struct tty_struct *tty, int break_state)
{
	struct linux_port_hwconf *info = (struct linux_port_hwconf *)tty->driver_data;
	unsigned long	flag;

	spin_lock_irqsave(&info->slock, flag);

	if (break_state == -1)
  {
		outb(inb(info->base + UART_LCR) |  UART_LCR_SBC, info->base + UART_LCR);
  }
	else
  {
		outb(inb(info->base + UART_LCR) & ~UART_LCR_SBC, info->base + UART_LCR);
  }

	spin_unlock_irqrestore(&info->slock,flag);
}


static inline void linux_modem_status_intr(struct linux_port_hwconf *info, unsigned char msr)
{
	if (msr & UART_MSR_TERI)
  {
	  info->icount.rng++;
  }
  
	if (msr & UART_MSR_DDSR)
  {
	  info->icount.dsr++;
  }
  
	if (msr & UART_MSR_DDCD)
  {
	  info->icount.dcd++;
  }
  
	if (msr & UART_MSR_DCTS)
  {
	  info->icount.cts++;
  }

	wake_up_interruptible(&info->delta_msr_wait);

	if ((info->flags & ASYNC_CHECK_CD) && (msr & UART_MSR_DDCD))
	{
	 	if (msr & UART_MSR_DCD)
    {
			wake_up_interruptible(&info->open_wait);
    }
		else if (!((info->flags & ASYNC_CALLOUT_ACTIVE) && (info->flags & ASYNC_CALLOUT_NOHUP)))
    {
     	set_bit(LINUX24_EVENT_HANGUP,&info->event);
    }

		MOD_INC_USE_COUNT;
		if (schedule_task(&info->tqueue) == 0)
    {
	 		MOD_DEC_USE_COUNT;
    }
	}

	if (info->flags & LINUX24_FLOWCTL_HW_AUTO)
	{
		if (info->tty->hw_stopped)
		{
			if (msr & UART_MSR_CTS)
			{
		   	info->tty->hw_stopped = 0;

				if ((info->type != PORT_16650) && (!info->ChipFlag))
				{
					info->IER |= UART_IER_THRI;
					outb(info->IER, info->base + UART_IER);
		  	}
				set_bit(LINUX24_EVENT_TXLOW,&info->event);

				MOD_INC_USE_COUNT;
				if (schedule_task(&info->tqueue) == 0)
        {
   	 			MOD_DEC_USE_COUNT;
        }
     	}
	  }
		else
		{
			if (!(msr & UART_MSR_CTS))
			{
				info->tty->hw_stopped = 1;
		    if ((info->type != PORT_16650) && (!info->ChipFlag))
				{
					info->IER &= ~UART_IER_THRI;
					outb(info->IER, info->base + UART_IER);
		    }
			}
	  }
	}
}


static inline void linux_data_transmit(struct linux_port_hwconf *info)
{
	int	count, cnt, tail_tmp = 0;
	unsigned char	tmp;
	struct tty_struct	*tty;
	unsigned long	flag;

	tty = info->tty;
	spin_lock_irqsave(&info->slock, flag);

	if ((info->tx_buf == 0))
	{
		spin_unlock_irqrestore(&info->slock,flag);
   	return;
	}

	if ((info->tx_buf_cont<=0) || (info->tty->stopped) || (tty->hw_stopped))
	{
		info->IER &= ~UART_IER_THRI;
		outb(info->IER, info->base + UART_IER);
		spin_unlock_irqrestore(&info->slock,flag);
		return;
	}

	cnt   = info->tx_buf_cont;
	count = info->xmit_fifo_size;

	do
	{
		tail_tmp = info->tx_buf_tail;
		tmp = info->tx_buf[info->tx_buf_tail++];
		outb(tmp, info->base + UART_TX);
   	info->tx_buf_tail = info->tx_buf_tail & (SERIAL_XMIT_SIZE - 1);

   	if ((--info->tx_buf_cont <= 0))
		{
			break;
		}

	} while (--count > 0);

	info->icount.tx += (cnt - info->tx_buf_cont);

 	if (info->tx_buf_cont < WAKEUP_CHARS)
	{
 		set_bit(LINUX24_EVENT_TXLOW, &info->event);

		MOD_INC_USE_COUNT;
  	if (schedule_task(&info->tqueue) == 0)
    {
 			MOD_DEC_USE_COUNT;
    }
 	}

	if (info->tx_buf_cont <= 0)
	{
		info->IER &= ~UART_IER_THRI;
		outb(info->IER, info->base + UART_IER);
	}

	spin_unlock_irqrestore(&info->slock,flag);
}


static inline void linux_data_receive(struct linux_port_hwconf *info, int *status, unsigned char iir)
{
	struct tty_struct *tty = info->tty;
	unsigned char	ch = 0x00, lsr = 0x00;
	int	count, ignored = 0, recv_room = 0;
	unsigned long	flag;
 	unsigned char *cp;
 	char 					*fp;

	recv_room = tty->ldisc.receive_room(tty);

 	if ((recv_room == 0) && (!info->ldisc_stop_rx))
	{
 		linux_throttle(tty);
		return;
 	}

	spin_lock_irqsave(&info->slock, flag);

 	cp = tty->flip.char_buf;
 	fp = tty->flip.flag_buf;
  count = 0;

	if (iir == UART_IIR_RLSI)
  {
		goto RLSI_START;
  } 

	if ((iir == UART_IIR_RDI) || (iir == UART_IIR_CTO))
	{
		do
		{
			ch = inb(info->base + UART_RX);
			info->icount.rx++;

			if (info->FC_flag & LINUX24_FLOWCTL_SW)
			{
				if (ch == STOP_CHAR(info->tty))
				{
					info->tty->stopped = 1;
 					info->IER |= UART_IER_THRI;
 					outb(info->IER, info->base + UART_IER);
					goto END_SWF;
				}

				if (ch == START_CHAR(info->tty))
				{
					info->tty->stopped = 0;
 					info->IER |= UART_IER_THRI;
 					outb(info->IER, info->base + UART_IER);
					goto END_SWF;
				}
			}

			*cp++ = ch;
 			*fp++ = 0;
			count++;
END_SWF:
			if ((iir == UART_IIR_RDI) && (count < 15))
      {
				lsr = 0x01;
      }
			else
      {
				lsr = inb(info->base + UART_LSR);
      }

		} while (lsr & UART_LSR_DR);
	}
	goto RECE_END;


RLSI_START:

	do
	{
   	ch = inb(info->base + UART_RX);
		info->icount.rx++;

		if (*status & (UART_LSR_BI | UART_LSR_PE | UART_LSR_FE | UART_LSR_OE))
		{
			if (*status & info->ignore_status_mask)
			{
				if (++ignored > 100)
        {
					break;
        }
				goto IGNORE_CHAR;
			}

			*status &= info->read_status_mask;

			if (*status & UART_LSR_OE)
			{
				*fp++ = TTY_OVERRUN;
				info->icount.overrun++;
				outb((info->FCR | 0x23), info->base + UART_FCR );
			}
			else if (*status & UART_LSR_BI)
			{
				info->icount.brk++;
				*fp++ = TTY_BREAK;
        
				if (info->flags & ASYNC_SAK)
        {
					do_SAK(tty);
        }
			}
			else if (*status & UART_LSR_PE)
			{
				*fp++ = TTY_PARITY;
				info->icount.parity++;
			}
			else if (*status & UART_LSR_FE)
			{
				*fp++ = TTY_FRAME;
				info->icount.frame++;
			}
			else
      {
				*fp++ = 0;
      }
		}

    
		*cp++ = ch;
		count++;
IGNORE_CHAR:
		*status = inb(info->base+UART_LSR);
	} while (*status & UART_LSR_DR);

RECE_END:

	if (count > 0)
  {
		tty->ldisc.receive_buf(tty, tty->flip.char_buf, tty->flip.flag_buf, count);
  }
  
	spin_unlock_irqrestore(&info->slock,flag);
}


static void linux_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
	struct linux_port_hwconf *info, *first_port;
	unsigned char msr;
	int	status, iir, i, max, irqbits, bits, pass_counter = 0, handled = 0;

	first_port = NULL;

  for (i = 0; i < LINUX24_MAX_BOARDS; i++)
	{
   	if (dev_id == &(linux_port_table[i * LINUX24_MAX_PORTS_PER_BOARD]))
		{
     	first_port = dev_id;
			break;
   	}
  }

  if (i == LINUX24_MAX_BOARDS)
  {
   	goto INTR_END;
  }

  if (first_port == 0)
  {
   	goto INTR_END;
  }

  max = linux_numports[linux_pci_board[i].board_type-1];

  // add by CHL 2006/10/02
	if ((first_port->remap) && ((first_port->subdeviceID == SUBDEVICE_ID_LINUX24_4027D) ||
				                      (first_port->subdeviceID == SUBDEVICE_ID_LINUX24_4037D) ||
                  				    (first_port->subdeviceID == SUBDEVICE_ID_LINUX24_4056D)))
  {
		while (1)
		{

      handled = 1;

			for (i = 0; i < LINUX24_MAX_PORTS_PER_BOARD; i++)
			{
				info = first_port + i;

				if (!info->base)
        {
					continue;
        }

				iir = inb(info->base + UART_IIR) & 0x0f;

				if (iir & UART_IIR_NO_INT)
        {
					continue;
        }

				status = inb(info->base + UART_LSR);

				if (iir == UART_IIR_RLSI || iir == UART_IIR_CTO || iir == UART_IIR_RDI)
        {
					linux_data_receive(info, &status, iir);
        }

				if ((iir == UART_IIR_THRI) && (status & UART_LSR_THRE))
        {
					linux_data_transmit(info);
        }

 				msr = inb(info->base + UART_MSR);

				if (msr & UART_MSR_ANY_DELTA)
        {
 	  			linux_modem_status_intr(info, msr);
        }
        else // add by CHL 2006/10/12, for 1699, modem interrupt can't be cleared
        {
          if ((iir == 0x00) && (info->ChipFlag == LINUX24_1699_HWID))
          {
		        info->IER |= UART_IER_THRI;
		        outb(info->IER, info->base + UART_IER);
          }
        }                   
			}

	    if (pass_counter++ > 128)
      {
			  break;
      }
		}
  }
  else
  {
	  while (1)
	  {
 		  irqbits = inb(first_port->vector) & first_port->vector_mask;

 		  if (irqbits == 0x00)
      {
			  break;
      }

 		  handled = 1;
     
 		  for (i = 0, bits = 1; i < max; i++, bits <<= 1)
		  {
        
			  if (!(bits & irqbits))
        {
				  continue;
        }

			  info = first_port + i;

			  iir = inb(info->base + UART_IIR) & 0x0f;
			  if (iir & UART_IIR_NO_INT)
        {
				  continue;
        }

			  status = inb(info->base + UART_LSR);

			  if (iir == UART_IIR_RLSI || iir == UART_IIR_CTO || iir == UART_IIR_RDI)
        {
				  linux_data_receive(info, &status, iir);
        }

			  if ((iir == UART_IIR_THRI) && (status & UART_LSR_THRE))
        {
				  linux_data_transmit(info);
        }

        
 				msr = inb(info->base + UART_MSR);

				if (msr & UART_MSR_ANY_DELTA)
        {
 	  			linux_modem_status_intr(info, msr);
        }
        else // add by CHL 2006/10/12, for 1699, modem interrupt can't be cleared
        {
          if ((iir == 0x00) && (info->ChipFlag == LINUX24_1699_HWID))
          {
		        info->IER |= UART_IER_THRI;
		        outb(info->IER, info->base + UART_IER);
          }
        }          
		  }

 		  if (pass_counter++ > 128)
      {
			  break;
      }
	  }
  }

INTR_END:
 	handled = 1;
}


static void linux_send_break(struct linux_port_hwconf *info, int duration)
{
	unsigned long	flag;

	if (!info->base)
  {
		return;
  }
  
	set_current_state(TASK_INTERRUPTIBLE);

	spin_lock_irqsave(&info->slock, flag);

	outb(inb(info->base + UART_LCR) | UART_LCR_SBC, info->base + UART_LCR);
	schedule_timeout(duration);
	outb(inb(info->base + UART_LCR) & ~UART_LCR_SBC, info->base + UART_LCR);

	spin_unlock_irqrestore(&info->slock,flag);
}


static int linux_get_modem_info(struct linux_port_hwconf *info, unsigned int *value)
{
	unsigned char	control, status;
	unsigned int result;
	unsigned long	flag;

	spin_lock_irqsave(&info->slock, flag);
	control = info->MCR;

	status=0;

	status = inb(info->base + UART_MSR);

	linux_msr[info->port] &= 0x0F;
	linux_msr[info->port] |= status;
	status = linux_msr[info->port];

 	spin_unlock_irqrestore(&info->slock,flag);

	if (status & UART_MSR_ANY_DELTA)
  {
    linux_modem_status_intr(info, status);
  }

 	result = 	((control & UART_MCR_RTS) ? TIOCM_RTS : 0) |
		 				((control & UART_MCR_DTR) ? TIOCM_DTR : 0) |
					  ((status  & UART_MSR_DCD) ? TIOCM_CAR : 0) |
		 				((status  & UART_MSR_RI)  ? TIOCM_RNG : 0) |
		 				((status  & UART_MSR_DSR) ? TIOCM_DSR : 0) |
		 				((status  & UART_MSR_CTS) ? TIOCM_CTS : 0);

	put_user(result, (unsigned long *)value);
	return(0);
}


static int linux_set_modem_info(struct linux_port_hwconf *info, unsigned int cmd, unsigned int *value)
{
	int error;
	unsigned int arg;

	error = verify_area(VERIFY_READ, value, sizeof(int));

	if (error)
  {
    return(error);
  }

	get_user(arg,(unsigned long *)value);

	switch (cmd)
	{
		case TIOCMBIS:
	    		if (arg & TIOCM_RTS)
          {
						info->MCR |= UART_MCR_RTS;
          }
          
	    		if (arg & TIOCM_DTR)
          {
						info->MCR |= UART_MCR_DTR;
          }
	    		break;
      
		case TIOCMBIC:
	    		if (arg & TIOCM_RTS)
          {
						info->MCR &= ~UART_MCR_RTS;
          }
            
	    		if (arg & TIOCM_DTR)
          {
						info->MCR &= ~UART_MCR_DTR;
          }
	    		break;

		case TIOCMSET:
					info->MCR &= ~(UART_MCR_RTS|UART_MCR_DTR);
					info->MCR  = ((arg & TIOCM_RTS)?UART_MCR_RTS:0) | ((arg & TIOCM_DTR)?UART_MCR_DTR:0);
	    		break;

		default:
	    		return(-EINVAL);
	}

	outb(info->MCR, info->base + UART_MCR);
	return(0);
}


static int linux_get_uart_info(struct linux_port_hwconf *info, struct serial_struct *retinfo)
{
	struct serial_struct tmp;

	if (!retinfo)
  {
    return(-EFAULT);
  }


	memset(&tmp, 0, sizeof(tmp));

	tmp.type 						= info->type;
	tmp.line 						= info->port;
	tmp.port 						= info->base;
	tmp.irq  						= info->irq;
	tmp.flags			 			= info->flags;
	tmp.baud_base 			= info->baud_base;
	tmp.close_delay 		= info->close_delay;
	tmp.closing_wait 		= info->closing_wait;
	tmp.custom_divisor 	= info->custom_divisor;
	tmp.hub6 						= 0;

	if (copy_to_user(retinfo, &tmp, sizeof(*retinfo)))
  {
		return -EFAULT;
  }

	return 0;
}


static int linux_set_uart_info(struct linux_port_hwconf *info, struct serial_struct *new_set)
{
	struct serial_struct local_serial;
	unsigned int flags;
	int retval = 0;


	if (!new_set || !info->base)
  {
		return(-EFAULT);
  }

	if (copy_from_user(&local_serial, new_set, sizeof(local_serial)))
  {
	  return -EFAULT;
  }

	if ((local_serial.irq != info->irq) ||
	    (local_serial.port != info->base) ||
	    (local_serial.custom_divisor != info->custom_divisor) ||
	    (local_serial.baud_base != info->baud_base))
  {
		return(-EPERM);
  }

	flags = info->flags & ASYNC_SPD_MASK;

  if (!capable(CAP_SYS_ADMIN))
	{
		if (( local_serial.baud_base != info->baud_base) ||
				( local_serial.close_delay != info->close_delay) ||
		 		((local_serial.flags & ~ASYNC_USR_MASK) !=
		 		(info->flags & ~ASYNC_USR_MASK)))
    {
			return(-EPERM);
    }
    
	  info->flags = ((info->flags & ~ASYNC_USR_MASK) | (local_serial.flags & ASYNC_USR_MASK));
	}
	else
	{
   	info->flags = ((info->flags & ~ASYNC_FLAGS) | (local_serial.flags & ASYNC_FLAGS));
	  info->close_delay = local_serial.close_delay * HZ/100;
	  info->closing_wait = local_serial.closing_wait * HZ/100;
	  info->tty->low_latency = (info->flags & ASYNC_LOW_LATENCY) ? 1 : 0;
	  info->type = local_serial.type;
  }

	if (info->ChipFlag == LINUX24_1889_HWID)
	{
		info->xmit_fifo_size	= 32;
		info->rx_trigger	= 16;
	}
	else if (info->ChipFlag == LINUX24_1699_HWID)
	{
		info->xmit_fifo_size	= 64;
		info->rx_trigger	= 32;
	}
	else
	{
		info->xmit_fifo_size	= 1;
		info->rx_trigger	= 1;
	}


	if (info->flags & ASYNC_INITIALIZED)
	{
	  if (flags != (info->flags & ASYNC_SPD_MASK))
		{
			termios_setup(info,0);
	  }
	}
	else
	{
	  retval = linux_startup(info);
	}

	return(retval);
}


static int linux_get_lsr_info(struct linux_port_hwconf *info, unsigned int *value)
{
	unsigned char	lsr;
	unsigned long	flag;

	spin_lock_irqsave(&info->slock, flag);
	lsr = inb(info->base + UART_LSR);
	spin_unlock_irqrestore(&info->slock,flag);

	if (copy_to_user(value, &lsr, sizeof(int)))
  {
		return -EFAULT;
  }

	return 0;
}


static void linux_set_termios(struct tty_struct *tty, struct termios *old_termios)
{
	struct linux_port_hwconf *info = (struct linux_port_hwconf *)tty->driver_data;

	if ((tty->termios->c_cflag != old_termios->c_cflag)||
		  (RELEVANT_IFLAG(tty->termios->c_iflag)!=RELEVANT_IFLAG(old_termios->c_iflag)))
	{
		termios_setup(info, old_termios);
	}

	if ((old_termios->c_cflag & CRTSCTS) && !(tty->termios->c_cflag & CRTSCTS))
	{
 	  tty->hw_stopped = 0;
		linux_start(tty);
	}

	if ((old_termios->c_iflag & IXON) && !(tty->termios->c_iflag & IXON))
	{
 		tty->stopped = 0;
    linux_start(tty);
	}
}


static int linux_ioctl(struct tty_struct *tty, struct file *file, unsigned int cmd, unsigned long arg)
{
 	struct linux_port_hwconf *info = (struct linux_port_hwconf *)tty->driver_data;
 	struct async_icount	cprev, cnow;
 	struct serial_icounter_struct	icount;
	struct port_info dump;

 	int	retval,tmp,error;
	unsigned long	flag;



	if ((cmd != TIOCGSERIAL) && (cmd != TIOCMIWAIT) && (cmd != TIOCGICOUNT))
	{
		if (tty->flags & (1<<TTY_IO_ERROR))
		{
			return(-EIO);
		}
	}

	switch (cmd)
	{

		case TCSBRK:
    		retval = tty_check_change(tty);
	   		if (retval)
        {
					return(retval);
        }

	   		tty_wait_until_sent(tty, 0);

				if (signal_pending(current))
        {
					return -EINTR;
        }

	   		if (!arg)
				{
					linux_send_break(info, HZ/4);
					if (signal_pending(current))
          {
						return -EINTR;
          }
				}
				return 0;

		case TCSBRKP:
		   	retval = tty_check_change(tty);
	  	 	if (retval)
        {
					return(retval);
        }

		   	tty_wait_until_sent(tty, 0);

				if (signal_pending(current))
        {
					return -EINTR;
        }

		   	linux_send_break(info, arg ? arg*(HZ/10) : HZ/4);
				if (signal_pending(current))
        {
					return -EINTR;
        }
  			return 0;

		case TIOCGSOFTCAR:
				tmp = C_CLOCAL(tty) ? 1 : 0;
				if (copy_to_user((void *)arg, &tmp, sizeof(int)))
        {
					return -EFAULT;
        }
	  		return 0;

		case TIOCSSOFTCAR:
				if (copy_from_user(&tmp, (void *)arg, sizeof(int)))
        {
					return -EFAULT;
        }
				tty->termios->c_cflag =	((tty->termios->c_cflag & ~CLOCAL) | (tmp ? CLOCAL : 0));
				return 0;

		case TIOCMGET:
		    error = verify_area(VERIFY_WRITE, (void *)arg, sizeof(unsigned int));

	  	  if (error)
        {
					return(error);
        }
		    return(linux_get_modem_info(info, (unsigned int *)arg));

		case TIOCMBIS:
		case TIOCMBIC:
		case TIOCMSET:
  	  	return(linux_set_modem_info(info, cmd, (unsigned int *)arg));

		case TIOCGSERIAL:
				return linux_get_uart_info(info, (struct serial_struct *) arg);

		case TIOCSSERIAL:
				return linux_set_uart_info(info, (struct serial_struct *) arg);

		case TIOCSERGETLSR:
				return linux_get_lsr_info(info, (unsigned int *) arg);


		case TIOCMIWAIT:
		{
 		  save_flags(flag);
			cli();

			cprev = info->icount;
			restore_flags(flag);

			info->IER |= UART_IER_MSI;
 			outb(info->IER, info->base + UART_IER);

			while (1)
			{
			  interruptible_sleep_on(&info->delta_msr_wait);
				if (signal_pending(current))
        {
				  return -ERESTARTSYS;
        }
				save_flags(flag);
				cli();

				cnow = info->icount;
				restore_flags(flag);

				if (cnow.rng == cprev.rng && cnow.dsr == cprev.dsr && cnow.dcd == cprev.dcd && cnow.cts == cprev.cts)
        {
					return -EIO;
        }  

				if (((arg & TIOCM_RNG) && (cnow.rng != cprev.rng)) ||
				    ((arg & TIOCM_DSR) && (cnow.dsr != cprev.dsr)) ||
				    ((arg & TIOCM_CD)  && (cnow.dcd != cprev.dcd)) ||
				    ((arg & TIOCM_CTS) && (cnow.cts != cprev.cts)))
				{
					return 0;
				}
				cprev = cnow;
			}
		}

		case TIOCGICOUNT:
				spin_lock_irqsave(&info->slock, flag);
				cnow = info->icount;
				spin_unlock_irqrestore(&info->slock,flag);
				icount.cts         = cnow.cts;
				icount.dsr         = cnow.dsr;
				icount.rng         = cnow.rng;
				icount.dcd         = cnow.dcd;
				icount.rx          = cnow.rx;
				icount.tx          = cnow.tx;
				icount.frame       = cnow.frame;
				icount.overrun     = cnow.overrun;
				icount.parity 	   = cnow.parity;
				icount.brk         = cnow.brk;
				icount.buf_overrun = cnow.buf_overrun;

				if (copy_to_user((void *)arg, &icount, sizeof(icount)))
        {
					return -EFAULT;
        }
				return 0;

		case LINUX24_DUMP_PORT_INFO:
				dump.board_name_info 	= info->board_name;
				dump.bus_number_info 	= info->bus_number;
				dump.dev_number_info 	= info->dev_number;
				dump.port_info 				= info->port;
				dump.base_info				= info->base;
				dump.irq_info					= info->irq;

				if (copy_to_user((void *)arg, &dump, sizeof(dump)))
        {
					return -EFAULT;
        }
				return 0;

		default:
	  	  return(-ENOIOCTLCMD);
	}

	return 0;
}


static void linux_do_softint(void *private_)
{
	struct linux_port_hwconf *info = (struct linux_port_hwconf *)private_;
	struct tty_struct *tty;

	tty = info->tty;

	if (tty)
	{
		if (test_and_clear_bit(LINUX24_EVENT_TXLOW, &info->event))
		{
     	if ((tty->flags &  (1 << TTY_DO_WRITE_WAKEUP)) && tty->ldisc.write_wakeup)
      {
				(tty->ldisc.write_wakeup)(tty);
      }

     	wake_up_interruptible(&tty->write_wait);
 		}

   	if (test_and_clear_bit(LINUX24_EVENT_HANGUP, &info->event))
    {
			tty_hangup(tty);
    }
	}

	MOD_DEC_USE_COUNT;
}


static int linux_get_PCI_conf(int board_type, struct linux_board_hwconf *hwconf )
{
	struct pci_dev *pdev = hwconf->pdev;
	int	i;
	unsigned int ioaddress[LINUX24_MAX_PORTS_PER_BOARD];


	hwconf->board_type = board_type;
	hwconf->ports      = linux_numports[board_type-1];

  // add by CHL 2006/10/02
	if ((hwconf->subdeviceID == SUBDEVICE_ID_LINUX24_4027D) ||
	    (hwconf->subdeviceID == SUBDEVICE_ID_LINUX24_4037D) ||
	    (hwconf->subdeviceID == SUBDEVICE_ID_LINUX24_4056D))
	{
		hwconf->remap = 1;
	}
	else
	{
		hwconf->remap = 0;
	}


  // add by CHL 2006/10/02
	if (hwconf->remap)
	{
		for (i = 0; i < hwconf->ports; i++)
		{
			if (i == 0)
			{
				ioaddress[i] = pci_resource_start(pdev, i+2);
				hwconf->ChipFlag[i] = LINUX24_1699_HWID;
			}
			else if (i == 1)
			{
				ioaddress[i] = pci_resource_start(pdev, i+2);
				hwconf->ChipFlag[i] = LINUX24_1699_HWID;
			}
			else if (i == 2)
			{
				ioaddress[i] = pci_resource_start(pdev, i+2);
				hwconf->ChipFlag[i] = LINUX24_1699_HWID;
			}
			else if (i == 3)
			{
				ioaddress[i] = pci_resource_start(pdev, i+2);
				hwconf->ChipFlag[i] = LINUX24_1699_HWID;
			}
		}
	}
  else
  {          
	  for (i = 0; i < hwconf->ports; i++)
	  {
		  if (i == 0)
		  {
			  ioaddress[i] = pci_resource_start(pdev, i);
			  hwconf->ChipFlag[i] = LINUX24_1889_HWID;
		  }
      else if (i == 1)
		  {
			  ioaddress[i] = ioaddress[i-1]+8;
			  hwconf->ChipFlag[i] = LINUX24_1889_HWID;
		  }
      else if (i == 2)
		  {
			  ioaddress[i] = pci_resource_start(pdev, i-1);
			  hwconf->ChipFlag[i] = LINUX24_1699_HWID;
		  }
      else if (i == 3)
		  {
			  ioaddress[i] = ioaddress[i-1]+8;
			  hwconf->ChipFlag[i] = LINUX24_1699_HWID;
		  }
      else if ((i == 4)||(i == 5)||(i == 6)||(i == 7))
		  {
			  ioaddress[i] = pci_resource_start(pdev, i-2);
			  hwconf->ChipFlag[i] = LINUX24_1699_HWID;
		  }
	  }
  }


	for (i = 0; i <  hwconf->ports; i++)
  {
		request_region(ioaddress[i],8, "linux24");
  }

	hwconf->vector_mask = 0;


	for (i = 0; i < hwconf->ports; i++)
	{
		hwconf->vector_mask  |= (1 << i);
		hwconf->ioaddr[i] = ioaddress[i];
		hwconf->baud_base[i] = 921600;
	}

	hwconf->vector 		= ioaddress[0] + 0x1c;
	hwconf->irq 			= hwconf->pdev->irq;
	hwconf->uart_type = PORT_16650;

	return(0);
}


static int linux_board_init(int board,struct linux_board_hwconf *hwconf)
{
	struct linux_port_hwconf *info;
	int retval, i, n;

	n = (board * LINUX24_MAX_PORTS_PER_BOARD);
	info = &linux_port_table[n];


	for (i = 0; i < hwconf->ports; i++, n++, info++)
	{
		info->board_name 			= hwconf->board_name;
		info->board_type			= hwconf->board_type;
		info->FC_flag     		= 0;
		info->bus_number			= hwconf->bus_number;
		info->dev_number			= hwconf->dev_number;
		info->count						= 0;
		info->port 						= n;
		info->base 						= hwconf->ioaddr[i];
		info->irq 						= hwconf->irq;
		info->vector 					= hwconf->vector;
		info->vector_mask 		= hwconf->vector_mask;
		info->ldisc_stop_rx		= 0;
		info->ChipFlag 				= hwconf->ChipFlag[i];
		info->flags 					= ASYNC_SHARE_IRQ;
		info->type 						= hwconf->uart_type;
		info->baud_base 			= hwconf->baud_base[i];
		info->custom_divisor 	= hwconf->baud_base[i] * 16;
		info->close_delay 		= 5*HZ/10;
		info->closing_wait 		= 30*HZ;
		info->err_shadow			= 0;

		info->normal_termios	= linux_tty_sdriver.init_termios;
		info->callout_termios	= linux_tty_cdriver.init_termios;

		info->tqueue.routine	= linux_do_softint;
		info->tqueue.data			= info;

		init_waitqueue_head(&info->open_wait);
		init_waitqueue_head(&info->close_wait);
		init_waitqueue_head(&info->delta_msr_wait);

		spin_lock_init(&info->slock);

		if (info->ChipFlag == LINUX24_1889_HWID)
		{
			info->xmit_fifo_size	= 32;
			info->rx_trigger	= 16;
		}
		else if (info->ChipFlag == LINUX24_1699_HWID)
		{
			info->xmit_fifo_size	= 32;
			info->rx_trigger	= 16;
		}
		else
		{
			info->xmit_fifo_size	= 1;
			info->rx_trigger	= 1;
		}

    //add by CHL 2006/10/02
		info->remap		    = hwconf->remap;
		info->subdeviceID	= hwconf->subdeviceID;      
	}

	for (i = 0; i < hwconf->ports; i++)
  {
		outb(inb(hwconf->ioaddr[i] + UART_IER) & 0xf0, hwconf->ioaddr[i]+UART_IER);
  }

	n = (board * LINUX24_MAX_PORTS_PER_BOARD);
	info = &linux_port_table[n];

	retval = request_irq(hwconf->irq, linux_interrupt, SA_SHIRQ, "linux24", info);
	if (retval)
	{
    printk("  Board %d: %d series", board, linux_board_name[hwconf->board_type-1]);
		printk("  Request irq fail,IRQ (%d) may be conflit with another device.\n",hwconf->irq);
		return(retval);
	}

  return 0;
}




int init_module(void)
{
	int	retval, ret1, ret2, i, index, b=0, m=0, n=0;
	unsigned short int subdeviceID;
	struct pci_dev *pdev=NULL;
	struct pci_dev *pdev_array[4] = {0,0,0,0};
  
	struct linux_board_hwconf hwconf;

  // change by CHL 2006/11/23
	printk("Loading PCI I/O Driver Module, Version: %s\n",DRIVER_VERSION);	

	spin_lock_init(&linux_io_lock);

	for (i = 0; i < LINUX24_MAX_BOARDS; i++)
	{
		linux_pci_board[i].board_type = -1;
	}


	memset(&linux_tty_sdriver,0,sizeof(struct tty_driver));

	linux_tty_sdriver.magic 								= TTY_DRIVER_MAGIC;
	linux_tty_sdriver.name 									= "ttyLNX";
	linux_tty_sdriver.major 								= LINUX24_TTY_MAJOR;
	linux_tty_sdriver.minor_start 					= 0;
	linux_tty_sdriver.num 									= LINUX24_MAX_PORTS_TOTAL;
	linux_tty_sdriver.type 									= TTY_DRIVER_TYPE_SERIAL,
	linux_tty_sdriver.subtype 							= SERIAL_TYPE_NORMAL,
	linux_tty_sdriver.flags 								= TTY_DRIVER_REAL_RAW,
	linux_tty_sdriver.init_termios 					= tty_std_termios;
	linux_tty_sdriver.init_termios.c_cflag 	= B9600 | CS8 | CREAD | HUPCL | CLOCAL;
  linux_tty_sdriver.init_termios.c_iflag 	= 0;

	linux_tty_sdriver.refcount 							= &linux_refcount;
	linux_tty_sdriver.table									= linux_tty;

	linux_tty_sdriver.termios								= linux_termios;
	linux_tty_sdriver.termios_locked				= linux_termios_locked;

	linux_tty_sdriver.open									= linux_open;
	linux_tty_sdriver.close									= linux_close;
	linux_tty_sdriver.write									= linux_write;
	linux_tty_sdriver.put_char							= linux_put_char;
	linux_tty_sdriver.flush_chars						= linux_flush_chars;
	linux_tty_sdriver.write_room						= linux_write_room;
	linux_tty_sdriver.chars_in_buffer				= linux_chars_in_buffer;
	linux_tty_sdriver.flush_buffer					= linux_flush_buffer;
	linux_tty_sdriver.ioctl									= linux_ioctl;
	linux_tty_sdriver.throttle							= linux_throttle;
	linux_tty_sdriver.unthrottle						= linux_unthrottle;
	linux_tty_sdriver.set_termios						= linux_set_termios;
	linux_tty_sdriver.stop									= linux_stop;
	linux_tty_sdriver.start									= linux_start;
	linux_tty_sdriver.hangup								= linux_hangup;

	linux_tty_sdriver.break_ctl 						= linux_rs_break;
	linux_tty_sdriver.wait_until_sent				= linux_wait_until_sent;

	linux_tty_cdriver												= linux_tty_sdriver;
	linux_tty_cdriver.name									= "cux";
	linux_tty_cdriver.major									= LINUX24_CAL_MAJOR;
	linux_tty_cdriver.subtype								= SERIAL_TYPE_CALLOUT;


	memset(linux_port_table, 0, LINUX24_MAX_PORTS_TOTAL * sizeof(struct linux_port_hwconf));
	memset(&hwconf, 0, sizeof(struct linux_board_hwconf));
	memset(&linux_msr, 0, LINUX24_MAX_PORTS_TOTAL * sizeof(unsigned char));

	n = (sizeof(linux_pcibrds) / sizeof(linux_pcibrds[0])) - 1;
	index = 0;
	b = 0;
	while (b < n)
	{
			pdev = pci_find_device(VENDOR_ID_LINUX24,DEVICE_ID_LINUX24, pdev);

			if (pdev == NULL)
			{
				b++;
				continue;
			}

 			if (m >= LINUX24_MAX_BOARDS)
			{
 				printk("PCI I/O Board Driver Module Support Four Boards In Maximum\n");
 				break;
 			}
			else
			{
				if ((b > 0) && ((pdev == pdev_array[0]) ||	(pdev == pdev_array[1]) ||	(pdev == pdev_array[2]) || (pdev == pdev_array[3])))
        {
					continue;
        }

				pci_read_config_word(pdev,0x2e,&subdeviceID);
				if (subdeviceID != linux_pcibrds[b].subdevice)
        {
					continue;
        }

				if (pci_enable_device(pdev))
				{
					printk("PCI Board Enable Fail !\n");
					continue;
				}
				pdev_array[m]      = pdev;
				hwconf.pdev        = pdev;
				hwconf.bus_number  = pdev->bus->number;
 				hwconf.dev_number  = PCI_SLOT(pdev->devfn);
				hwconf.board_name  = linux_board_name[(linux_pcibrds[b].driver_data)-1];
        hwconf.subdeviceID = subdeviceID; //add by CHL 2006/10/02
        
				printk("Found %d series PCI I/O Board , BUSNumber : %d  DEVNumber : %d\n",linux_board_name[(linux_pcibrds[b].driver_data)-1],
                                                                                  hwconf.bus_number,
                                                                                  hwconf.dev_number);
        
				retval = linux_get_PCI_conf((int)linux_pcibrds[b].driver_data,&hwconf);
				linux_pci_board[m] = hwconf;
   			linux_board_init(m,&hwconf);
				m++;
		  }
	}


	ret1 = ret2 = 0;
  if (!(ret1 = tty_register_driver(&linux_tty_sdriver)))
	{
		if (!(ret2 = tty_register_driver(&linux_tty_cdriver)))
		{
      printk("Loading PCI I/O Driver Module Done\n");
			return 0;
		}
		else
		{
			tty_unregister_driver(&linux_tty_sdriver);
			printk("Couldn't Load PCI I/O Board Driver Module\n");
		}
	}
  else
	{
		printk("Couldn't Load PCI I/O Driver Module\n");
	}



  if (ret1 || ret2)
	{
		for (i = 0; i < LINUX24_MAX_BOARDS; i++)
		{
     	if (linux_pci_board[i].board_type == -1)
      {
 		  	continue;
      }
      else
			{
		  	free_irq(linux_pci_board[i].irq, &linux_port_table[i*LINUX24_MAX_PORTS_PER_BOARD]);
			}
    }
    return -1;
  }

	printk("Loading PCI I/O Board Driver Module Done\n");
	return (0);
}


void cleanup_module(void)
{
	int	i, j, error=0;

	if ((error |= tty_unregister_driver(&linux_tty_cdriver)))
  {
		printk("Couldn't Unload PCI I/O Board driver\n");
  }

	if ((error |= tty_unregister_driver(&linux_tty_sdriver)))
  {
		printk("Couldn't Unload PCI I/O Board driver\n");
  }

  
	for (i = 0; i < LINUX24_MAX_BOARDS; i++)
	{
		struct pci_dev *pdev;

		if (linux_pci_board[i].board_type == -1)
    {
			continue;
    }
		else
		{
			pdev = linux_pci_board[i].pdev;

			free_irq(linux_pci_board[i].irq, &linux_port_table[i*LINUX24_MAX_PORTS_PER_BOARD]);
			if (pdev != NULL)
			{
				for (j = 0; j < linux_pci_board[i].ports; j++)
        {
					release_region(linux_pci_board[i].ioaddr[j], 8);
        }
			}
		}
	}

	printk("Unload PCI I/O Board Driver Module\n");
}


