/*
 *                                                                                     	
 *                                Linux Dumb Terminal                               
 *                                                                                      
 *                                                                     
 *                                                                                      
 *                                                              Version: 1.00           	
 *                                                              Date: 2006/07/10        	
 */
#include <stdio.h>
#include <stdlib.h>
#include <ncurses.h>
#include <form.h>
#include <termios.h>
#include <fcntl.h>
#include <string.h>
#include <pthread.h>
#include <limits.h>
#include "lnxterm.h"

static int n_mainMENU    = sizeof(mainTable) / sizeof(char *);
static int n_patnMENU    = sizeof(patnTable) / sizeof(char *);

static void windowsInit(void );
static void mainFunction(int );
static void printMainMenu_L(int );
static int  printMainMenu(int );

static void serialPortSetup(void );
static void printPortBasic(int , int );

static void commPamSetup(void );
static void printCommPam(void );
static void scanDeviceName(void );

static void sendPattern(void );
static int  printPatternSelect(int );
static void printPatternSelect_L(int );
static int  printDefaultPattern(void );
static int  arrangeUserDefine(void );
static int  scanUserDefine(void );

static void printErrMsg(int );
static void printMessage(void );

static void termioConfSetup(void );
static int  openDevice(void );
static void closeDevice(void );
static void *readTerm(void );
static void *sentTerm(void );
static void dumbTerm1(void );
static void dumbTerm2(void );


struct linux_term_conf 
{
	int			fd;
	struct 	termios		t;	
	struct	termios		back;	
	int			flag;
	
	int			flowctlChoice;	
	int			baudrateChoice;
	int			databitChoice;
	int			parityChoice;
	int			stopbitChoice;
	
};

static struct linux_term_conf	term;

struct handler_thread
{
	pthread_t 	thread;
	int		thr_id;
};

static struct handler_thread *r_thread;
static struct handler_thread *w_thread;


static int memory_patn = 1;
static int userDataLen = 0;
static long int count;

static WINDOW  	*menu_win;
static WINDOW	*port_win;														
static WINDOW	*comm_win;													
static WINDOW  	*mesg_win;
static WINDOW  	*term_win;
static WINDOW  	*eror_win;
static WINDOW	*patn_win;
static WINDOW	*scan_win;
static WINDOW	*daut_win;


int  main(void)
{
	memset(&term, 0, sizeof(struct linux_term_conf));
	term.t.c_cflag     = B9600 | CS8 | CLOCAL | CREAD | CRTSCTS;
	term.t.c_iflag     = IGNPAR;
 	term.t.c_cc[VMIN]  = 1;
 	term.t.c_cc[VTIME] = 0;
	term.flag          = FLAG_FIRSTOPEN | FLAG_TMIOFIRSTSET;
	 
	term.flowctlChoice  = 0;
	term.baudrateChoice = 4;
	term.databitChoice  = 3;
	term.parityChoice   = 0;
	term.stopbitChoice  = 0;
	
	initscr();
	noecho();
	cbreak();
	windowsInit();	
	mainFunction(1);	
	endwin();
	closeDevice();	
	return 0;
}

static void windowsInit(void)
{	
	menu_win = newwin(WIN_MAIN_H, WIN_MAIN_W, (WIN_TERM_H - WIN_MAIN_H)/2, (WIN_TERM_W - WIN_MAIN_W)/2);
	port_win = newwin(WIN_PORT_H, WIN_PORT_W, WIN_PORT_STR_Y, WIN_PORT_STR_X);	
	comm_win = newwin(WIN_COMM_H, WIN_COMM_W, WIN_COMM_STR_Y, WIN_COMM_STR_X);
	mesg_win = newwin(WIN_MESG_H, WIN_MESG_W, WIN_MESG_STR_Y, WIN_MESG_STR_X);
	term_win = newwin(WIN_DUMB_H, WIN_DUMB_W, WIN_DUMB_STR_Y, WIN_DUMB_STR_X);	
	eror_win = newwin(WIN_EROR_H, WIN_EROR_W, WIN_EROR_STR_Y, WIN_EROR_STR_X);
	patn_win = newwin(WIN_PATN_H, WIN_PATN_W, WIN_PATN_STR_Y, WIN_PATN_STR_X);
	scan_win = newwin(WIN_SCAN_H, WIN_SCAN_W, WIN_SCAN_STR_Y, WIN_SCAN_STR_X);	
	daut_win = newwin(WIN_DAUT_H, WIN_DAUT_W, WIN_DAUT_STR_Y, WIN_DAUT_STR_X);	
		
	keypad(menu_win, TRUE);	
	keypad(port_win, TRUE);
	keypad(comm_win, TRUE);
	keypad(term_win, TRUE);			
	keypad(eror_win, TRUE);
	keypad(patn_win, TRUE);	
	keypad(scan_win, TRUE);		
	keypad(daut_win, TRUE);	
}

static void mainFunction(int select)
{
	int mainMenuSlt = select, close = 0;
	int ret_patn = 1, i;

	while(1)
	{
		mainMenuSlt = printMainMenu(mainMenuSlt);

		switch(mainMenuSlt)
		{	
			case RET_EXITPROGRAM:
				close = 1;
				break;
	
			case RET_COMMSETUP :		
				serialPortSetup();
				break;
				
			case RET_DUMBTERM_ASCII :
				if(openDevice())
				{		
				}
				else
				{	
					term.flag |= FLAG_ASCIIMODE;	
					werase(menu_win);
					wrefresh(menu_win);						
					dumbTerm1();
				}
				mainMenuSlt = 2;
				break;	
			
			case RET_DUMBTERM_HEX :
				if(openDevice())
				{
				}
				else
				{	
					term.flag |= FLAG_HEXMODE;	
					werase(menu_win);
					wrefresh(menu_win);						
					dumbTerm1();
				}
				mainMenuSlt = 3;
				break;	
				
													
			case RET_SENDPATTERN :
				sendPattern();		
				mainMenuSlt = 4;
				break;	
		}
	
		if(close != 0)
			break;
	}



	free(r_thread);
	free(w_thread);
	clear();
	endwin();
	closeDevice();
}

static int printMainMenu(int select)
{
	int highlight = select, choice = 0, close = 0, ch;
	curs_set(0);	
	printMainMenu_L(highlight);
		
	while(1)
	{	
		ch = wgetch(menu_win);
		switch(ch)
		{	
			case KEY_UP:
				if(highlight == 1)
					highlight = n_mainMENU;
				else
					--highlight;
				break;
				
			case KEY_DOWN:
				if(highlight == n_mainMENU)
					highlight = 1;
				else 
					++highlight;
				break;
			
			case KEY_SESC:
				choice = RET_EXITPROGRAM;
				break;
				
			case KEY_SENTER:
				choice = highlight;
				break;
		}	

		printMainMenu_L(highlight);
		
		if(choice != 0)
			break;
	}
	
	curs_set(1);	
	return choice;		
}

static void printMainMenu_L(int select)
{
	int y = 1, i;	
	box(menu_win, 0, 0);
	mvwprintw(menu_win, 0,10,"[ main menu ]");

	for(i = 0; i < n_mainMENU; ++i)
	{	
		mvwprintw(menu_win, y, 1, "%s", mainTable[i]);
		++y;
	}

	y = 1;	
	for(i = 0; i < n_mainMENU; ++i)
	{	
		if(select == (i + 1)) 
		{	
			wattron(menu_win, A_REVERSE); 
			mvwprintw(menu_win, y, 1, "%s", mainTable[i]);		
			wattroff(menu_win, A_REVERSE);
		}
		++y;		
	}	
	
	wrefresh(menu_win);	
}

static void serialPortSetup(void)
{
	int ch, close = 0;
	printPortBasic(0,0);
	
	while(1)
	{
		ch = wgetch(port_win);
		switch(ch)
		{
		
			case 'a':
			case 'A':
				scanDeviceName();
				break;
				
			case 'b':
			case 'B':
				commPamSetup();
				printMainMenu_L(1);
				break;
				
			case 'c':
			case 'C':	
				term.flowctlChoice++;
				if(term.flowctlChoice > 2)
					term.flowctlChoice = 0;
					
				term.t.c_cflag &= ~CRTSCTS;
				term.t.c_iflag &= ~(IXON|IXOFF);
				

				if(term.flowctlChoice == 0)
				{
					term.t.c_cflag |=  CRTSCTS;
				}
				else if(term.flowctlChoice == 1)
				{
					term.t.c_iflag |= (IXON|IXOFF);
				}
				break;
				
			case KEY_SENTER:
				close = 1;
				break;
		}
				
		printPortBasic(0,0);

		if(close != 0)
			break;
	}	
	
	werase(port_win);
	wrefresh(port_win);
}

static void printPortBasic(int curX, int curY)
{
	int i, y;
	box(port_win, 0, 0);
	
	y = 1;
	for(i = 0; i < (sizeof(portSetupTable) / sizeof(char *)); i++)
	{	
		mvwprintw(port_win, y, 1, "%s", portSetupTable[i]);
		++y;
	}	
	
	y = 1;
	for(i = 0; i < 3; i++)
	{
		if(i==0)
		{
			mvwprintw(port_win, y, 22, "%s", deviceName);
		}
		else if(i == 1)
		{	
			mvwprintw(port_win, y, 22, "%s", baudTable[term.baudrateChoice]);
			wprintw(port_win, "%s", dataTable[term.databitChoice]);
			wprintw(port_win, "%s", pariTable[term.parityChoice]);
			wprintw(port_win, "%s", stopTable[term.stopbitChoice]);
		}
		else if(i == 2)
		{
			mvwprintw(port_win, y, 22, "%s", flowTable[term.flowctlChoice]);
		}
			
		y++;		
	}
	
	mvwprintw(port_win, 5, 2,"Select item, or <Enter> to exit? ");
	
	if((curX != 0) || (curY != 0))
		wmove(port_win, curY, curX);
	
	wrefresh(port_win);
	
}

static void commPamSetup(void)
{
	int ch,close = 0;
	printCommPam();
	
	while(1)
	{
		ch = wgetch(comm_win);
		switch(ch)
		{
			case 'a':
			case 'A':
				term.t.c_cflag &= ~(B300|B1200|B2400|B4800|B9600|B19200|B38400|B57600|B115200|B230400|B460800|B921600);
				term.t.c_cflag |= B300;
				term.baudrateChoice = 0;
				break;
				
			case 'b':
			case 'B':
				term.t.c_cflag &= ~(B300|B1200|B2400|B4800|B9600|B19200|B38400|B57600|B115200|B230400|B460800|B921600);
				term.t.c_cflag |= B1200;
				term.baudrateChoice = 1;
				break;
					
			case 'c':
			case 'C':
				term.t.c_cflag &= ~(B300|B1200|B2400|B4800|B9600|B19200|B38400|B57600|B115200|B230400|B460800|B921600);
				term.t.c_cflag |= B2400;
				term.baudrateChoice = 2;
				break;
				
			case 'd':
			case 'D':
				term.t.c_cflag &= ~(B300|B1200|B2400|B4800|B9600|B19200|B38400|B57600|B115200|B230400|B460800|B921600);
				term.t.c_cflag |= B4800;
				term.baudrateChoice = 3;
				break;
				
			case 'e':
			case 'E':
				term.t.c_cflag &= ~(B300|B1200|B2400|B4800|B9600|B19200|B38400|B57600|B115200|B230400|B460800|B921600);
				term.t.c_cflag |= B9600;
				term.baudrateChoice = 4;
				break;					
				
			case 'f':
			case 'F':
				term.t.c_cflag &= ~(B300|B1200|B2400|B4800|B9600|B19200|B38400|B57600|B115200|B230400|B460800|B921600);
				term.t.c_cflag |= B19200;
				term.baudrateChoice = 5;
				break;				
				
			case 'g':
			case 'G':
				term.t.c_cflag &= ~(B300|B1200|B2400|B4800|B9600|B19200|B38400|B57600|B115200|B230400|B460800|B921600);
				term.t.c_cflag |= B38400;
				term.baudrateChoice = 6;
				break;				
				
			case 'h':
			case 'H':
				term.t.c_cflag &= ~(B300|B1200|B2400|B4800|B9600|B19200|B38400|B57600|B115200|B230400|B460800|B921600);
				term.t.c_cflag |= B57600;
				term.baudrateChoice = 7;
				break;				
				
			case 'i':
			case 'I':
				term.t.c_cflag &= ~(B300|B1200|B2400|B4800|B9600|B19200|B38400|B57600|B115200|B230400|B460800|B921600);
				term.t.c_cflag |= B115200;
				term.baudrateChoice = 8;
				break;				
				
			case 'j':
			case 'J':
				term.t.c_cflag &= ~(B300|B1200|B2400|B4800|B9600|B19200|B38400|B57600|B115200|B230400|B460800|B921600);
				term.t.c_cflag |= B230400;
				term.baudrateChoice = 9;
				break;				
				
			case 'k':
			case 'K':
				term.t.c_cflag &= ~(B300|B1200|B2400|B4800|B9600|B19200|B38400|B57600|B115200|B230400|B460800|B921600);
				term.t.c_cflag |= B460800;
				term.baudrateChoice = 10;
				break;				
				
			case 'l':
			case 'L':
				term.t.c_cflag &= ~(B300|B1200|B2400|B4800|B9600|B19200|B38400|B57600|B115200|B230400|B460800|B921600);
				term.t.c_cflag |= B921600;
				term.baudrateChoice = 11;
				break;				
				
			case 'm':
			case 'M':
				term.t.c_cflag &= ~(PARENB|PARODD);			
				term.parityChoice = 0;
				break;		
				
			case 'n':
			case 'N':
				term.t.c_cflag &= ~(PARENB|PARODD);
				term.t.c_cflag |= PARENB;			
				term.parityChoice = 1;
				break;	
				
			case 'o':
			case 'O':
				term.t.c_cflag &= ~(PARENB|PARODD);
				term.t.c_cflag |= (PARENB|PARODD);			
				term.parityChoice = 2;
				break;									
				
			case 'p':
			case 'P':
				term.t.c_cflag &= ~CSTOPB;			
				term.stopbitChoice = 0;
				break;	
				
			case 'q':
			case 'Q':
				term.t.c_cflag |= CSTOPB;	
				term.stopbitChoice = 1;
				break;					
				
			case 'r':
			case 'R':
				term.t.c_cflag &= ~(CS5|CS6|CS7|CS8);
				term.t.c_cflag |= CS5;
				term.databitChoice = 0;
				break;					
				
			case 's':
			case 'S':
				term.t.c_cflag &= ~(CS5|CS6|CS7|CS8);
				term.t.c_cflag |= CS6;
				term.databitChoice = 1;
				break;					
				
			case 't':
			case 'T':
				term.t.c_cflag &= ~(CS5|CS6|CS7|CS8);
				term.t.c_cflag |= CS7;
				term.databitChoice = 2;
				break;					
				
			case 'u':
			case 'U':
				term.t.c_cflag &= ~(CS5|CS6|CS7|CS8);
				term.t.c_cflag |= CS8;
				term.databitChoice = 3;
				break;					
														
			case KEY_SENTER:
				close = 1;
				break;	
		}
		
		printCommPam();
		
		if(close != 0)
			break;
	}	
	
	werase(comm_win);
	wrefresh(comm_win);
}

static void printCommPam(void)
{
	int i, y, ch;

	box(comm_win, 0, 0);
	mvwprintw(comm_win, 0,10,"[ comm param setup ]");	
	mvwprintw(comm_win, 2, 2," Current: %s%s%s%s"	,baudTable[term.baudrateChoice]
							,dataTable[term.databitChoice]
							,pariTable[term.parityChoice]
							,stopTable[term.stopbitChoice]);
							
	mvwprintw(comm_win, 4, 2," Baud rate    Parity    Data bit   ");
	mvwprintw(comm_win, 6, 2," A: 300       M: None   R: 5       ");
	mvwprintw(comm_win, 7, 2," B: 1200      N: Even   S: 6       ");						
	mvwprintw(comm_win, 8, 2," C: 2400      O: Odd    T: 7       ");
	mvwprintw(comm_win, 9, 2," D: 4800                U: 8       ");
	mvwprintw(comm_win,10, 2," E: 9600                           ");
	mvwprintw(comm_win,11, 2," F: 19200                          ");
	mvwprintw(comm_win,12, 2," G: 38400     Stop bit             ");
	mvwprintw(comm_win,13, 2," H: 57600                          ");
	mvwprintw(comm_win,14, 2," I: 115200    P: 1                 ");
	mvwprintw(comm_win,15, 2," J: 230400    Q: 2                 ");
	mvwprintw(comm_win,16, 2," K: 460800                         ");
	mvwprintw(comm_win,17, 2," L: 921600                         ");
	
	mvwprintw(comm_win,20, 2," Choice, or <Enter> to exit? ");
	wrefresh(comm_win);
}

static void scanDeviceName(void)
{
	int ch, curX, curY, close = 0, i;
	
	curX = 22;
	curY = 1;
	wmove(port_win, curY,curX);
	
	while(1)
	{
		ch = wgetch(port_win);
		switch(ch)
		{
			case KEY_LEFT:
				curX--;	
				if(curX < 22)
					curX = 36;		
				break;
			
					
			case KEY_RIGHT:
				curX++;
				if(curX > 36)
					curX = 22;								
				break;	
		
				
			case KEY_SENTER:
				close = 1;				
				break;	
				
			case KEY_SBACKSPACE:
				getyx(port_win, curY, curX);
					
				if(curX == 22)
					deviceName[14] = ' ';
				else									
					deviceName[curX - 23] = ' ';		
					
				curX--;
				if(curX < 22)
					curX = 36;
	
				break;
				
			case KEY_SDEL:
				getyx(port_win, curY, curX);
				for(i=(curX - 22); i<15; i++)
				{
					if(i >= 14)
						deviceName[i] = ' ';
					else	
						deviceName[i] = deviceName[i+1];
				}
				break;	
								
			default:
				if((ch >= 0x20) && (ch <= 0x7E))
				{
					getyx(port_win, curY, curX);
					deviceName[curX - 22] = ch;
					curX++;
					if(curX > 36)
						curX = 22;
				}								
		}
		
		printPortBasic(curX,curY);	
		
		if(close != 0)
			break;
	}		
}

static void sendPattern(void)
{
	int close = 0, ret_patn, i;
	

	while(1)
	{		

		ret_patn = printPatternSelect(memory_patn);
		if(ret_patn == 0)
		{
			close = 1;
		}	
		else if(ret_patn == 1)
		{
			term.flag |= FLAG_PATTERNMODE;	
			
			if(printDefaultPattern() == 1)
			{
				memory_patn = 1;				
			}
			else
			{
				werase(patn_win);
				werase(menu_win);
				
				wrefresh(patn_win);
				wrefresh(menu_win);
			
			
				bzero(userData,sizeof(userData));					
				for(i=0;i<40;i++)
					userData[i] = '0'+i;
		
				userDataLen = 40;
				dumbTerm2();
				close = 1;
			}
		}
		else if(ret_patn == 2)
		{
			term.flag |= FLAG_PATTERNMODE;	
			userDataLen = scanUserDefine();
			if(!userDataLen)
			{
				memory_patn = 2;
			}
			else
			{
				werase(patn_win);
				werase(menu_win);
				
				wrefresh(patn_win);
				wrefresh(menu_win);
							
				dumbTerm2();
				close = 1;
			}
		}
		
		
		
		if(close != 0)
			break;	
	}
}

static int printPatternSelect(int select)
{
	int highlight = select, choice = 0, close = 0, ch;
	printMainMenu_L(4);
	printPatternSelect_L(highlight);
	
	while(1)
	{	
		curs_set(0);
		ch = wgetch(patn_win);
		
		switch(ch)
		{	
			case KEY_UP:
				if(highlight == 1)
					highlight = n_patnMENU;
				else
					--highlight;
				break;
				
			case KEY_DOWN:
				if(highlight == n_patnMENU)
					highlight = 1;
				else 
					++highlight;
				break;
			
			case KEY_SESC:
				close = 1;
				break;
				
			case KEY_SENTER:
				if(openDevice())
					close = 1;					
				else
					choice = highlight;
								
				break;
		}
		
		curs_set(1);
		
		
		printPatternSelect_L(highlight);
		
		if(close != 0)
		{
			memory_patn = highlight;
			choice = 0;
			werase(patn_win);
			wrefresh(patn_win);			
			break;
		}
		
		if(choice != 0)
			break;
	}
	
	return choice;
}

static void printPatternSelect_L(int select)
{
	int y = 1, i;	
	box(patn_win, 0, 0);
	mvwprintw(patn_win, 0, 7,"[ pattern menu ]");
	
	for(i = 0; i < n_patnMENU; ++i)
	{	
		mvwprintw(patn_win, y, 1, "%s", patnTable[i]);
		++y;
	}

	y = 1;	
	for(i = 0; i < n_patnMENU; ++i)
	{	
		if(select == i + 1) 
		{	
			wattron(patn_win, A_REVERSE); 
			mvwprintw(patn_win, y, 1, "%s", patnTable[i]);		
			wattroff(patn_win, A_REVERSE);
		}
		++y;		
	}	
	
	wrefresh(patn_win);	
}

static int printDefaultPattern(void)
{
	int ch, status = 0;
	box(daut_win, 0, 0);
	curs_set(0);		 
	
	mvwprintw(daut_win,0, 9,"[ default defined pattern ]");
	mvwprintw(daut_win, 5, 1,"  <Enter> to send         <Esc> to exit  ");	
	wattron(daut_win, A_REVERSE); 
	mvwprintw(daut_win, 2, 2,"0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVW");	
	wattroff(daut_win, A_REVERSE);		
	wrefresh(daut_win);

	while(1)
	{	
		ch = wgetch(daut_win);
		switch(ch)
		{		
			case KEY_SESC:
				status = 1;
				break;
				
			case KEY_SENTER:	
				status = 2;	
				break;
		}

		if(status != 0)
			break;
	}
	
	curs_set(1);
	werase(daut_win);
	wrefresh(daut_win);
	return status;
}

static void printMessage(void)
{	
	if(term.flag & FLAG_ASCIIMODE)
		mvwprintw(mesg_win,0,0,"Sending and Receiving chars (ASCII Mode)                Ctrl-N: main menu");
	else if(term.flag & FLAG_HEXMODE)
		mvwprintw(mesg_win,0,0,"Sending and Receiving chars (Hex Mode)                  Ctrl-N: main menu");	
	else if(term.flag & FLAG_PATTERNMODE)
		mvwprintw(mesg_win,0,0,"Sending and receiving patterns                          Ctrl-N: main menu");		
	
		
	mvwprintw(mesg_win,1, 0,"Device=%s,",fd);
		
	mvwprintw(mesg_win,1,22,"Setting=%s,%s,%s,%s"	,baudTable[term.baudrateChoice]
							,dataTable[term.databitChoice]
							,pariTable[term.parityChoice]
							,stopTable[term.stopbitChoice]);		
							
	mvwhline(mesg_win,3, 0, ACS_HLINE, WIN_TERM_W);
	wrefresh(mesg_win);
}

static int scanUserDefine(void)
{
	int y = 3, i, close, curX, curY;
	int ch;
	box(scan_win, 0, 0);	
	
	mvwprintw(scan_win, 0, 18,"[ user defined pattern ]");
	mvwhline(scan_win, 3, 18, ACS_HLINE, 42); 
	mvwprintw(scan_win, 5, 1,"  <Enter> to send   <Ctrl-d> change mode   <Esc> to exit  ");	
	mvwprintw(scan_win, 2, 1,"  String (ab..): ");
	wrefresh(scan_win);
	
	bzero(userDefine,sizeof(userDefine));
	for(i=0 ;i<42; i++)
		userDefine[i]=' ';	
		
	mvwprintw(scan_win,2,18,"%s",userDefine);
	wmove(scan_win, 2,18);		
	getyx(scan_win, curY, curX);
	
	term.flag |= FLAG_SCANSTRING;
	
	while(1)
	{	
		ch = wgetch(scan_win);
		switch(ch)
		{			
			case KEY_SESC:
				werase(scan_win);
				wrefresh(scan_win);			
				return 0;
				break;
				
			case KEY_SENTER:	
				if(openDevice())
				{
					werase(scan_win);
					wrefresh(scan_win);					
					return 0;
				}	
				else
				{
					userDataLen = arrangeUserDefine();
					if(!userDataLen)
						break;
					else
					{
						werase(scan_win);
						wrefresh(scan_win);
						return userDataLen;
					}
				}
				break;
					
			case KEY_LEFT:	
				getyx(scan_win, curY, curX);	
				if(term.flag & FLAG_SCANSTRING)
				{
					curX--;	
					if(curX < 18)
						curX = 59;
				}
				else
				{
					curX--;
					if(!((curX+1) % 3))
						curX--;
					
					if(curX < 18)
						curX = 58;					
				}			
				wmove(scan_win, curY , curX);
				wrefresh(scan_win);											
				break;
				
			case KEY_RIGHT:
				getyx(scan_win, curY, curX);
				if(term.flag & FLAG_SCANSTRING)
				{			
					curX++;

					if(curX > 59)
						curX = 18;
				}
				else
				{
					curX++;
					if(!((curX+1) % 3))
						curX++;

					if(curX > 59)
						curX = 18;
				}		
				wmove(scan_win, curY , curX);				
				wrefresh(scan_win);								
				break;
				
			case KEY_CTL_D:
				if(term.flag & FLAG_SCANSTRING)
				{
					term.flag &= ~FLAG_SCANSTRING;	
					term.flag |=  FLAG_SCANASCII;			
					mvwprintw(scan_win, 2, 1,"  Hex (0b,1f,.): ");
					
					for(i=0; i<42; i++)
					{
						if(!((i+1) % 3))
							userDefine[i]=',';
						else
							userDefine[i]=' ';
					}	
				}
				else
				{
					term.flag &= ~FLAG_SCANASCII;
					term.flag |=  FLAG_SCANSTRING;	
					mvwprintw(scan_win, 2, 1,"  String (ab..): ");		
					for(i=0 ;i<42; i++)
						userDefine[i]=' ';	
				}
				mvwprintw(scan_win,2,18,"%s",userDefine);
				wmove(scan_win, 2,18);
				wrefresh(scan_win);					
				break;	
			
			case KEY_SBACKSPACE:
				if(term.flag & FLAG_SCANSTRING)
				{			
					getyx(scan_win, curY, curX);
					if(curX == 18)
						userDefine[41] = ' ';
					else									
						userDefine[curX - 19] = ' ';		

					curX--;
					if(curX < 18)
						curX = 59;
				}
				else if(term.flag & FLAG_SCANASCII)
				{
					getyx(scan_win, curY, curX);
					if( (!(curX % 3)) && (curX!=18) )
						curX--;
						
					if(curX == 18)
						userDefine[40] = ' ';
					else									
						userDefine[curX - 19] = ' ';		

					curX--;
					if(curX < 18)
						curX = 58;					
				}
				mvwprintw(scan_win,2,18,"%s",userDefine);
				wmove(scan_win, curY,curX);
				wrefresh(scan_win);
				break;
					
			default:
				if(term.flag & FLAG_SCANSTRING)
				{				
					if((ch >= 0x20) && (ch <= 0x7E))
					{
						getyx(scan_win, curY, curX);							
						userDefine[curX - 18] = ch;
						curX++;
						
						if(curX > 59)
							curX = 18;						
					}						
				}
				else if(term.flag & FLAG_SCANASCII)
				{
					if( ((ch>='0') && (ch<='9')) || ((ch>='a') && (ch<='f')) )
					{
						getyx(scan_win, curY, curX);
						userDefine[curX - 18] = ch;
						curX++;
								
						if(!((curX+1) % 3))
							curX++;

						if(curX > 59)
							curX = 18;					
					}				
				}
				mvwprintw(scan_win,2,18,"%s",userDefine);
				wrefresh(scan_win);
				wmove(scan_win, curY , curX);
				wrefresh(scan_win);						
				break;
		}
	}
}

static int arrangeUserDefine(void)
{
	int i,j, numToTrans;
	unsigned char num[2];
	bzero(num,sizeof(num));
	bzero(userData,sizeof(userData));
	
	if(term.flag & FLAG_SCANSTRING)
	{
		j=0;
		for(i=0; i<42; i++)
		{
			if(userDefine[i]!=' ')
			{
				userData[j] = userDefine[i];
				j++;
			}  	
		}
	}
	else if(term.flag & FLAG_SCANASCII)
	{
		j=0;
		for(i=0; i<14; i++)
		{
			if( (userDefine[i+(i*2)]!=' ') || (userDefine[i+(i*2)+1]!=' ') )
			{
				num[0] = userDefine[i+(i*2)];
				num[1] = userDefine[i+(i*2)+1];								 
				userData[j] = toascii(strtol(num,NULL,16));
				
				if((userDefine[i+(i*2)] >= 'a') && (userDefine[i+(i*2)+1]!=' '))
					userData[j] |= 0x80; 					

				j++;	
			}		
		}
	}
	return j;
}

static void termioConfSetup(void)
{
	if((term.t.c_cflag != term.back.c_cflag) || (term.t.c_iflag != term.back.c_iflag))
		term.flag |= FLAG_TMIOCHANGE;
	
	if((term.flag & FLAG_TMIOCHANGE) || (term.flag & FLAG_TMIOFIRSTSET))
	{
		term.flag &= ~(FLAG_TMIOCHANGE | FLAG_TMIOFIRSTSET);	
		tcflush(term.fd, TCIOFLUSH);
		tcsetattr(term.fd, TCSANOW, &term.t);
		term.back = term.t;
	}	
}

static void printErrMsg(int errType)
{
	curs_set(0);	
	box(eror_win, 0, 0);

	if(errType == ERRTYPE_OPEN)
	{
		mvwprintw(eror_win, 0,10,"[error msg]");
		mvwprintw(eror_win, 1, 2,"%s",fd);
		mvwprintw(eror_win, 1,19,"open fail");
		wrefresh(eror_win);
	}
	
	wgetch(eror_win);	
	werase(eror_win);
	wrefresh(eror_win);
	curs_set(1);
}

static int openDevice(void)
{
	int i,hasChar = 0;
	bzero(&fd,sizeof(fd));
	
	for(i=0; i<15; i++)
	{
		if((deviceName[14 - i] == ' ') && (hasChar == 0))
			continue;
		else
		{			
			hasChar = 1;	
			fd[14-i] = deviceName[14-i];
		}
	}

	
	if(term.flag & FLAG_OPENED)
	{
		for(i=0; i<15; i++)
		{
			if(deviceName[i] != deviceNameBack[i])
				term.flag |= FLAG_OPENCHANGE;
		}		
	}

	
	if((term.flag & FLAG_OPENCHANGE) || (term.flag & FLAG_FIRSTOPEN) || (term.flag & FLAG_OPENERR))
	{	
		term.flag &= ~(FLAG_OPENCHANGE | FLAG_FIRSTOPEN | FLAG_OPENERR);
		closeDevice();
		term.fd = open(fd, O_RDWR | O_NOCTTY);
		if(term.fd < 0)
		{
			printErrMsg(ERRTYPE_OPEN);
			term.flag |= FLAG_OPENERR;
			return RETURN_ERROR;	
		}
	
		term.flag |= (FLAG_OPENED | FLAG_TMIOFIRSTSET);			
		termioConfSetup();
		
		for(i=0; i<15; i++)
			deviceNameBack[i] = deviceName[i];	
		
		return RETURN_OK;		
	}
	else
	{
		termioConfSetup();
		return RETURN_OK;	
	}
}

static void closeDevice(void)
{
	if(term.fd > 0)
	{
		tcflush(term.fd, TCIOFLUSH);
		close(term.fd);
	}
}

static void *readTerm(void)
{
	unsigned char ch[RECVBUFFSIZE];
	int number, i, curX = 0, curY = 0;
	int j,k;
	
	pthread_setcancelstate(PTHREAD_CANCEL_ENABLE,NULL);
	pthread_setcanceltype(PTHREAD_CANCEL_DEFERRED,NULL);
	
	wmove(term_win, curY , curX);
	wrefresh(term_win);
	
	scrollok(term_win,TRUE);
	while(1)
	{
		bzero(ch,sizeof(ch));
		
		number = read(term.fd, &ch, RECVBUFFSIZE);		
		if(number > 0)
		{		
			if(term.flag & (FLAG_ASCIIMODE | FLAG_PATTERNMODE))
			{
				for(i = 0; i < number; i++)
				{
					mvwprintw(term_win, curY, curX,"%c",ch[i]);
					if(curX > (WIN_DUMB_W - 2))
					{					
						curX = 0;
						curY++;
				
						if(curY > (WIN_DUMB_H - 1 ))
							curY = WIN_DUMB_H -1;
					}
					else
						curX++;					
				}				
			}
			else if(term.flag & FLAG_HEXMODE)
			{
				for(i = 0; i < number; i++)
				{				
					mvwprintw(term_win, curY, curX,"%2x",ch[i]);
					if(curX > (WIN_DUMB_W - 3))
					{						
						curX = 0;
						curY++;				
						if(curY > (WIN_DUMB_H - 1 ))
							curY = WIN_DUMB_H -1;
					}
					else
						curX = curX + 2;	
				}
			}
			
			wmove(term_win, curY, curX);
			wrefresh(term_win);				
		}
	}
	pthread_exit(NULL);
}

static void *sentTerm(void)
{
	int i, j;
	pthread_setcancelstate(PTHREAD_CANCEL_ENABLE,NULL);
	pthread_setcanceltype(PTHREAD_CANCEL_DEFERRED,NULL);
	//count = 0;
	//mvwprintw(mesg_win,2,0,"Count=%d                   ",count);
	//wrefresh(mesg_win);	
	curs_set(0);
	while(1)
	{	
		write(term.fd,&userData,userDataLen);	
		count++;
		
		for(i=0; i<100; i++)
		{
			for(j=0; j<1000; j++)
			{
			}
			
			/*
			if(i == 80)
			{
				mvwprintw(mesg_win,2,6,"%d                 ",count);
				wrefresh(mesg_win);
			}	
			*/
		}
	}
	curs_set(1);
	pthread_exit(NULL);
}

static void dumbTerm1(void)
{
	pthread_attr_t custom_attr;
	char ch;
	int close = 0;
	
	r_thread = (struct handler_thread *)malloc(sizeof(struct handler_thread));	
	pthread_attr_init(&custom_attr);
	pthread_attr_setstacksize(&custom_attr, PTHREAD_STACK_MIN);
	r_thread->thr_id = pthread_create(&r_thread->thread, &custom_attr, (void *)readTerm, (void *)NULL);
	
	printMessage();
	
	while(1)
	{	
		ch = wgetch(term_win);
		switch(ch)
		{	
			case KEY_CTL_N:
				pthread_cancel(r_thread->thread);
				close = 1;
				break;
			default:
				write(term.fd, &ch, 1);
				break;
		}
		
		if(close != 0)
			break;
	}
	
	pthread_join(r_thread->thread, NULL);
	term.flag &= ~(FLAG_ASCIIMODE | FLAG_HEXMODE);
	werase(mesg_win);	
	wrefresh(mesg_win);	
	werase(term_win);	
	wrefresh(term_win);
	tcflush(term.fd, TCIOFLUSH);	
}

static void dumbTerm2(void)
{
	pthread_attr_t custom_attr;
	char ch;
	int close = 0;

	w_thread = (struct handler_thread *)malloc(sizeof(struct handler_thread));
	r_thread = (struct handler_thread *)malloc(sizeof(struct handler_thread));

			
	pthread_attr_init(&custom_attr);
	pthread_attr_setstacksize(&custom_attr, PTHREAD_STACK_MIN);

	r_thread->thr_id = pthread_create(&r_thread->thread, &custom_attr, (void *)readTerm, (void *)NULL);
	w_thread->thr_id = pthread_create(&w_thread->thread, &custom_attr, (void *)sentTerm, (void *)NULL);
		
	printMessage();
	
	while(1)
	{	
		ch = wgetch(term_win);
		switch(ch)
		{	
			case KEY_CTL_N:
				pthread_cancel(r_thread->thread);
				pthread_cancel(w_thread->thread);
				close = 1;
				break;
		}
		
		
		if(close != 0)
			break;
	}
	

	pthread_detach(r_thread->thread);
	pthread_detach(w_thread->thread);
	term.flag &= ~(FLAG_ASCIIMODE | FLAG_HEXMODE | FLAG_PATTERNMODE| FLAG_SCANSTRING | FLAG_SCANASCII);
	werase(mesg_win);	
	wrefresh(mesg_win);	
	werase(term_win);	
	wrefresh(term_win);		
	tcflush(term.fd, TCIOFLUSH);		
}

