/****************************************************************************************/
/*                                                                                      */	
/*                              SUNIX PCI I/O Board Device Driver                       */
/*                                                                                      */
/*                                      For Kernel 2.6.x                                */
/*                                                                                      */
/*                                                              Version: 1.02           */	
/*                                                              Date: 07/10/2006        */	
/****************************************************************************************/
#include <linux/version.h>
#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/init.h>
#include <linux/module.h>
#include <linux/slab.h>
#include <linux/wait.h>
#include <linux/tty.h>
#include <linux/tty_driver.h>
#include <linux/tty_flip.h>
#include <linux/serial.h>
#include <linux/serial_reg.h>
#include <linux/mm.h>
#include <linux/signal.h>
#include <linux/pci.h>
#include <linux/workqueue.h>
#include <linux/interrupt.h>
#include <linux/circ_buf.h>

#include <asm/uaccess.h>
#include <asm/system.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/segment.h>
#include <asm/serial.h>

#include "sunix.h"

MODULE_AUTHOR(DRIVER_AUTHOR);
MODULE_DESCRIPTION(DRIVER_DESC);
MODULE_LICENSE("GPL");

enum	{
// 1S
 	SUNIX_BOARD_4027 = 1,	

// 2S
	SUNIX_BOARD_4036,
	SUNIX_BOARD_4037,
	SUNIX_BOARD_8139,

// 4S
	SUNIX_BOARD_4055,
	SUNIX_BOARD_4056,
	SUNIX_BOARD_8159,

// 8S
	SUNIX_BOARD_4066, 
	SUNIX_BOARD_8169,
};

static unsigned int sunix_board_name[] = 
{
// 1S
 	4027,	

// 2S
	4036,
	4037,
	8139,

// 4S
	4055,
	4056,
	8159,

// 8S
	4066,
	8169,
};

static int sunix_numports[] = 
{
// 1S
 	1,	// 4027

// 2S
	2,	// 4036
	2,	// 4037	
	2,	// 8139

// 4S
	4,	// 4055
	4,	// 4056
	4,	// 8159

// 8S
	8,	// 4066
	8,	// 8169


};


static  struct pci_device_id	sunix_pcibrds[] = 
{
// 1S
	{PCI_VENDOR_ID_SUNIX,PCI_DEVICE_ID_SUNIX ,PCI_SUBVENDOR_ID_SUNIX, PCI_SUBDEVICE_ID_SUNIX_4027A,  0, 0,SUNIX_BOARD_4027},	
	{PCI_VENDOR_ID_SUNIX,PCI_DEVICE_ID_SUNIX ,PCI_SUBVENDOR_ID_SUNIX, PCI_SUBDEVICE_ID_SUNIX_4027D,  0, 0,SUNIX_BOARD_4027},

// 2S
	{PCI_VENDOR_ID_SUNIX,PCI_DEVICE_ID_SUNIX ,PCI_SUBVENDOR_ID_SUNIX, PCI_SUBDEVICE_ID_SUNIX_4036A,  0, 0,SUNIX_BOARD_4036},
	{PCI_VENDOR_ID_SUNIX,PCI_DEVICE_ID_SUNIX ,PCI_SUBVENDOR_ID_SUNIX, PCI_SUBDEVICE_ID_SUNIX_4036D,  0, 0,SUNIX_BOARD_4036},
	{PCI_VENDOR_ID_SUNIX,PCI_DEVICE_ID_SUNIX ,PCI_SUBVENDOR_ID_SUNIX, PCI_SUBDEVICE_ID_SUNIX_4037A,  0, 0,SUNIX_BOARD_4037},
	{PCI_VENDOR_ID_SUNIX,PCI_DEVICE_ID_SUNIX ,PCI_SUBVENDOR_ID_SUNIX, PCI_SUBDEVICE_ID_SUNIX_4037D,  0, 0,SUNIX_BOARD_4037},
	{PCI_VENDOR_ID_SUNIX,PCI_DEVICE_ID_SUNIX ,PCI_SUBVENDOR_ID_SUNIX, PCI_SUBDEVICE_ID_SUNIX_8139,   0, 0,SUNIX_BOARD_8139},

// 4S
	{PCI_VENDOR_ID_SUNIX,PCI_DEVICE_ID_SUNIX ,PCI_SUBVENDOR_ID_SUNIX, PCI_SUBDEVICE_ID_SUNIX_4055WN, 0, 0,SUNIX_BOARD_4055},	
	{PCI_VENDOR_ID_SUNIX,PCI_DEVICE_ID_SUNIX ,PCI_SUBVENDOR_ID_SUNIX, PCI_SUBDEVICE_ID_SUNIX_4056A,  0, 0,SUNIX_BOARD_4056},
	{PCI_VENDOR_ID_SUNIX,PCI_DEVICE_ID_SUNIX ,PCI_SUBVENDOR_ID_SUNIX, PCI_SUBDEVICE_ID_SUNIX_4056D,  0, 0,SUNIX_BOARD_4056},
	{PCI_VENDOR_ID_SUNIX,PCI_DEVICE_ID_SUNIX ,PCI_SUBVENDOR_ID_SUNIX, PCI_SUBDEVICE_ID_SUNIX_8159,   0, 0,SUNIX_BOARD_8159},

// 8S
 	{PCI_VENDOR_ID_SUNIX,PCI_DEVICE_ID_SUNIX ,PCI_SUBVENDOR_ID_SUNIX, PCI_SUBDEVICE_ID_SUNIX_4066A,  0, 0,SUNIX_BOARD_4066},
 	{PCI_VENDOR_ID_SUNIX,PCI_DEVICE_ID_SUNIX ,PCI_SUBVENDOR_ID_SUNIX, PCI_SUBDEVICE_ID_SUNIX_4066R,  0, 0,SUNIX_BOARD_4066},
	{PCI_VENDOR_ID_SUNIX,PCI_DEVICE_ID_SUNIX ,PCI_SUBVENDOR_ID_SUNIX, PCI_SUBDEVICE_ID_SUNIX_8169,   0, 0,SUNIX_BOARD_8169},
	{0}
};
MODULE_DEVICE_TABLE(pci, sunix_pcibrds);


struct sunix_board_hwconf 
{
	unsigned int		board_name;
	unsigned int		bus_number;
	unsigned int		dev_number;
	int			board_type;
	int			ports;
	int			irq;
	int			vector;
	int			vector_mask;
	int			uart_type;	
	int			ioaddr[SUNIX_MAX_PORTS_PER_BOARD];
	struct 	pci_dev		*pdev;		
	int			SunixChipFlag[SUNIX_MAX_PORTS_PER_BOARD];		
	int			baud_base[SUNIX_MAX_PORTS_PER_BOARD];
};


struct sunix_port_hwconf 
{
	unsigned int		board_name;		
	int			board_type;
	int			baud_rate;	
	unsigned char		FC_flag;		
	unsigned int		bus_number;		
	unsigned int		dev_number;				
	int			port;			
	int			base;			
	int			irq;			
	int			vector;			
	int			vector_mask;		
	int			rx_trigger;		
	int			baud_base;		
	int			flags;	
	int			type;			
	struct tty_struct	*tty;	
	int			read_status_mask;
	int			ignore_status_mask;
	int			xmit_fifo_size;	
	int			custom_divisor;	
	unsigned char		x_char; 			
	int			close_delay;
	unsigned short		closing_wait;
	unsigned char		IER;			
	unsigned char		MCR;			
	unsigned char		FCR;
	unsigned long		event;	
	int			count;	
	int			blocked_open;	
	long			session;	
	long			pgrp;		
	unsigned char		*tx_buf;
	int			tx_buf_cont;	
	int			tx_buf_head;
	int			tx_buf_tail;

	struct work_struct 	tqueue;
	struct termios		normal_termios;
	wait_queue_head_t 	open_wait;
	wait_queue_head_t 	close_wait;
	wait_queue_head_t 	delta_msr_wait;	
	struct async_icount	icount; 
 	int			timeout;
	int			SunixChipFlag;	
 	unsigned char		ldisc_stop_rx;		
	unsigned char           err_shadow;	
	spinlock_t		slock;		
};


static struct tty_driver		*sunix_tty_driver;
static struct tty_struct		*sunix_tty[SUNIX_MAX_PORTS_TOTAL];
static struct termios			*sunix_termios[SUNIX_MAX_PORTS_TOTAL];
static struct termios			*sunix_termios_locked[SUNIX_MAX_PORTS_TOTAL];
static spinlock_t			 sunix_io_lock;
static struct sunix_board_hwconf	 sunix_pci_board[SUNIX_MAX_BOARDS];
static struct sunix_port_hwconf		 sunix_port_table[SUNIX_MAX_PORTS_TOTAL + 1];



static int  baudrate_setup(struct sunix_port_hwconf *, unsigned int );
static int  termios_setup(struct sunix_port_hwconf *, struct termios *);
static int  sunix_startup(struct sunix_port_hwconf *);
static int  sunix_block_til_ready(struct tty_struct *, struct file *, struct sunix_port_hwconf *);
static void sunix_shutdown(struct sunix_port_hwconf *);
static void sunix_wait_until_sent(struct tty_struct *, int );
static void sunix_flush_chars(struct tty_struct *);
static int  sunix_open(struct tty_struct *, struct file *);
static void sunix_close(struct tty_struct *, struct file *);
static void sunix_put_char(struct tty_struct *, unsigned char );
static int  sunix_write_room(struct tty_struct *);
static int  sunix_chars_in_buffer(struct tty_struct *);
static void sunix_flush_buffer(struct tty_struct *);
static void sunix_throttle(struct tty_struct *);
static void sunix_unthrottle(struct tty_struct *);
static void sunix_stop(struct tty_struct *);
static void sunix_start(struct tty_struct *);
static void sunix_hangup(struct tty_struct *);
static void sunix_rs_break(struct tty_struct *, int );
static void sunix_send_break(struct sunix_port_hwconf *, int );
static int  sunix_tiocmget(struct tty_struct *, struct file *);
static int  sunix_tiocmset(struct tty_struct *, struct file *,unsigned int , unsigned int );
static int  sunix_get_uart_info(struct sunix_port_hwconf *, struct serial_struct *);
static int  sunix_set_uart_info(struct sunix_port_hwconf *, struct serial_struct *);
static int  sunix_get_lsr_info(struct sunix_port_hwconf *in, unsigned int *);
static void sunix_set_termios(struct tty_struct *, struct termios *);
static int  sunix_ioctl(struct tty_struct * , struct file *,unsigned int , unsigned long );
static void sunix_do_softint(void *);
static int  sunix_get_PCI_conf(int , struct sunix_board_hwconf *);
static int  sunix_board_init(int ,struct sunix_board_hwconf *);

static irqreturn_t sunix_interrupt(int , void *, struct pt_regs *);

#if (LINUX_VERSION_CODE > KERNEL_VERSION(2,6,10))
static int  sunix_write(struct tty_struct *, const unsigned char *, int );
#else
static int  sunix_write(struct tty_struct *, int , const unsigned char *, int );
#endif


#if (LINUX_VERSION_CODE > KERNEL_VERSION(2,6,10))
static void sunix_modem_status_intr(struct sunix_port_hwconf *, unsigned char );
#else
static inline void sunix_modem_status_intr(struct sunix_port_hwconf *, unsigned char );
#endif

#if (LINUX_VERSION_CODE > KERNEL_VERSION(2,6,10))
static void sunix_data_transmit(struct sunix_port_hwconf *);
#else
static inline void sunix_data_transmit(struct sunix_port_hwconf *);
#endif

#if (LINUX_VERSION_CODE > KERNEL_VERSION(2,6,10))
static void sunix_data_receive(struct sunix_port_hwconf *, int *, unsigned char );
#else
static inline void sunix_data_receive(struct sunix_port_hwconf *, int *, unsigned char );
#endif














static int baudrate_setup(struct sunix_port_hwconf *info, unsigned int new_speed)
{
	int dlldlm = 0, retval = 0;
	unsigned char lcr;

	if (!info->tty || !info->tty->termios) 
		return retval;
		
	if (!info->base) 
		return retval;

   	info->baud_rate = new_speed;
	
	 if (new_speed) 
	 {
	    	dlldlm = (info->baud_base * 16) / (new_speed * 16);
	  	if( dlldlm == 0 )
	        	dlldlm = 1;
	} 
	else 
	    	dlldlm = 0;

	
	info->timeout = ((info->xmit_fifo_size * HZ * 10 * dlldlm) / info->baud_base);
	
		
	if (dlldlm) 
	{
	
    		info->MCR |= UART_MCR_DTR;
    		outb(info->MCR, info->base + UART_MCR);
	}
	else 
	{
    		info->MCR &= ~UART_MCR_DTR;
    		outb(info->MCR, info->base + UART_MCR);
    		return retval;
	}

    	lcr = inb(info->base + UART_LCR);

	outb(lcr | UART_LCR_DLAB, info->base + UART_LCR);	
	outb(dlldlm & 0xff, info->base + UART_DLL);
	outb(dlldlm >> 8  , info->base + UART_DLM); 
	outb(lcr, info->base + UART_LCR);	
	return retval;
}


static int termios_setup(struct sunix_port_hwconf *info, struct termios *old_termios)
{
	unsigned int c_cflag, baudrate = 0;
	unsigned char lcr = 0;
        int retval = 0;

	if (!info->tty || !info->tty->termios) 
		return retval;

	c_cflag = info->tty->termios->c_cflag;

	if (!info->base) 
		return retval;

	switch(c_cflag & (CBAUD | CBAUDEX))
	{
        	case B921600	: baudrate = 921600;	break;
        	case B460800	: baudrate = 460800;	break;
        	case B230400	: baudrate = 230400;	break;
        	case B115200	: baudrate = 115200;	break;
        	case B57600	: baudrate = 57600;	break;
        	case B38400	: baudrate = 38400;	break;
        	case B19200	: baudrate = 19200;	break;
        	case B9600	: baudrate = 9600;	break;
        	case B4800	: baudrate = 4800;	break;
        	case B2400	: baudrate = 2400;	break;
        	case B1200	: baudrate = 1200;	break;
        	case B300	: baudrate = 300;	break;
        	case B110	: baudrate = 110;	break;
        	default		: baudrate = 0;		break;
	}
		
       	baudrate_setup(info, baudrate);
 
	switch (c_cflag & CSIZE) 
	{
		case CS5: 	lcr = 0x00; break;
		case CS6: 	lcr = 0x01; break;
		case CS7: 	lcr = 0x02; break;
		case CS8: 	lcr = 0x03; break;
		default:  	lcr = 0x03; break;
	}
		
	if (c_cflag & CSTOPB) 
		lcr |= 0x04;

	if (c_cflag & PARENB) 
	{
		if (c_cflag & PARODD) 
			lcr |= 0x08;
		else 	
			lcr |= 0x18;
	}		

	if (c_cflag & CMSPAR)  
		lcr |= UART_LCR_SPAR;  

	
	info->FCR = UART_FCR_ENABLE_FIFO;
	
	if(info->SunixChipFlag == SUNIX_1889_HWID) 
	{
		switch (info->rx_trigger) 
		{
			case 8: 	info->FCR |= 0x21; break;
			case 16: 	info->FCR |= 0x61; break;
			case 24: 	info->FCR |= 0xa1; break;
			case 28: 	info->FCR |= 0xe1; break;
			default:  	info->FCR |= 0x61; break;
		}	
	}
	else if(info->SunixChipFlag  == SUNIX_1699_HWID) 
	{	
		switch (info->rx_trigger)
		{
			case 16: 	info->FCR |= 0x31; break;
			case 32: 	info->FCR |= 0x71; break;
			case 48: 	info->FCR |= 0xb1; break;
			case 56: 	info->FCR |= 0xf1; break;
			default:  	info->FCR |= 0x31; break;
		}		
	}	

	info->IER &= ~UART_IER_MSI;  
	info->MCR &= ~UART_MCR_AFE;  

	if (c_cflag & CRTSCTS) 
	{
		info->FC_flag |= SUNIX_FLOWCTL_HW_AUTO;
		info->IER     |= UART_IER_MSI;
 		info->MCR     |= UART_MCR_AFE;
	} 
	else 
		info->FC_flag &= ~SUNIX_FLOWCTL_HW_AUTO;		
				
	
		
	if (c_cflag & CLOCAL) 
		info->flags &= ~ASYNC_CHECK_CD;  
		
	else 
	{
		info->flags |= ASYNC_CHECK_CD;
		info->IER   |= UART_IER_MSI;
	}
		
	
	
	if (I_IXON(info->tty) || I_IXOFF(info->tty)) 
		info->FC_flag |= SUNIX_FLOWCTL_SW;			
	else  
		info->FC_flag &= ~SUNIX_FLOWCTL_SW;		
			

	info->read_status_mask = UART_LSR_DR | UART_LSR_OE | UART_LSR_THRE;
	
	if (I_INPCK(info->tty))   
		info->read_status_mask |= UART_LSR_FE | UART_LSR_PE;
		
		
	if (I_BRKINT(info->tty) || I_PARMRK(info->tty))  
	    	info->read_status_mask |= UART_LSR_BI;
		
		
	info->ignore_status_mask = 0;

	if (I_IGNBRK(info->tty)) 
	{ 
	    	info->ignore_status_mask |= UART_LSR_BI;
	    	info->read_status_mask   |= UART_LSR_BI;

	    	if (I_IGNPAR(info->tty)) 
		{   
			info->ignore_status_mask |= UART_LSR_OE | UART_LSR_PE | UART_LSR_FE;
			info->read_status_mask   |= UART_LSR_OE | UART_LSR_PE | UART_LSR_FE;
	    	}
	}
	
	outb(info->MCR, info->base + UART_MCR);	
	outb(info->IER, info->base + UART_IER);	

	outb(info->FCR, info->base + UART_FCR);	
	outb(lcr,  info->base + UART_LCR);  
	
	
        return retval;
	
}

static int sunix_startup(struct sunix_port_hwconf *info)
{
	unsigned long page;
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,11))	
	unsigned long flag;
#endif
	page = __get_free_page(GFP_KERNEL);
	
	if ( !page ) 
		return(-ENOMEM);

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,11))	
	spin_lock_irqsave(&info->slock, flag);
#endif
		
	if (info->flags & ASYNC_INITIALIZED) 
	{
	    	free_page(page);

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,11))	
		spin_unlock_irqrestore(&info->slock,flag);
#endif
	    	return(0);
	}

	if (!info->base || !info->type) 
	{
	    	if (info->tty)
			set_bit(TTY_IO_ERROR, &info->tty->flags);
			
	    	free_page(page);
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,11))	
		spin_unlock_irqrestore(&info->slock,flag);
#endif
	    	return(0);
	}
		
	if (info->tx_buf)
	    	free_page(page);
	else
	    	info->tx_buf = (unsigned char *)page;	
		
	
 	outb((UART_FCR_CLEAR_RCVR | UART_FCR_CLEAR_XMIT), info->base + UART_FCR);	

	
	if (inb(info->base + UART_LSR) == 0xff) 
	{
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,11))	
		spin_unlock_irqrestore(&info->slock,flag);
#endif		
    		if (capable(CAP_SYS_ADMIN)) 
		{
			if ( info->tty )
				set_bit(TTY_IO_ERROR, &info->tty->flags);
			return(0);
	   	}
		else
			return(-ENODEV);
	}	
	
	
 	(void)inb(info->base + UART_LSR);
 	(void)inb(info->base + UART_RX );
 	(void)inb(info->base + UART_IIR);
 	(void)inb(info->base + UART_MSR);
	
 	outb(UART_LCR_WLEN8, info->base + UART_LCR);	
	
 	info->MCR = UART_MCR_DTR | UART_MCR_RTS;
	outb(info->MCR, info->base + UART_MCR);
	
	info->IER = UART_IER_MSI | UART_IER_RLSI | UART_IER_RDI; 
	
  	if (info->SunixChipFlag  == SUNIX_1699_HWID)
  		info->IER |= SUNIX_1699_CLK_DIVIDER_DISABLE;
 		

 	outb(info->IER, info->base + UART_IER);	
	
 	(void)inb(info->base + UART_LSR);
 	(void)inb(info->base + UART_RX );
 	(void)inb(info->base + UART_IIR);
 	(void)inb(info->base + UART_MSR);

  	if (info->tty)
 		clear_bit(TTY_IO_ERROR, &info->tty->flags); 
	    
  	info->tx_buf_cont = info->tx_buf_head = info->tx_buf_tail = 0;


#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,11))	
	spin_unlock_irqrestore(&info->slock,flag);
#endif
    	termios_setup(info, 0);
	
	info->flags |= ASYNC_INITIALIZED;
	return 0;
}


static int sunix_block_til_ready(struct tty_struct *tty, struct file * filp, struct sunix_port_hwconf *info)
{
	DECLARE_WAITQUEUE(wait, current);
	int retval, do_clocal = 0;
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,11))	
	unsigned long flag;
#endif	
	if ((filp->f_flags & O_NONBLOCK) || (tty->flags & (1 <<TTY_IO_ERROR))) 
	{
	    	info->flags |= ASYNC_NORMAL_ACTIVE;
	    	return(0);
	}

    	if (tty->termios->c_cflag & CLOCAL)
		do_clocal = 1;
		
	retval = 0;
	add_wait_queue(&info->open_wait, &wait);

	
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,11))	
	spin_lock_irqsave(&info->slock, flag);
#endif	

	if (!tty_hung_up_p(filp)) 
		info->count--;
		

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,11))	
	spin_unlock_irqrestore(&info->slock,flag);
#endif


	info->blocked_open++;
	
	
	while (1) 
	{
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,11))	
		spin_lock_irqsave(&info->slock, flag);
#endif	
		outb(inb(info->base + UART_MCR) | UART_MCR_DTR | UART_MCR_RTS, info->base + UART_MCR);

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,11))	
		spin_unlock_irqrestore(&info->slock,flag);
#endif

	    	set_current_state(TASK_INTERRUPTIBLE);
			
	    	if (tty_hung_up_p(filp) || !(info->flags & ASYNC_INITIALIZED)) 
		{
			if (info->flags & ASYNC_HUP_NOTIFY) 
				retval = -EAGAIN;	
			else
		    		retval = -ERESTARTSYS;

			break;
	   	}
				
	    	if (!(info->flags & ASYNC_CLOSING) && (do_clocal || (inb(info->base + UART_MSR) & UART_MSR_DCD)))
			break;
				
	    	if (signal_pending(current)) 
		{
			retval = -ERESTARTSYS;
			break;
	    	}
				
	    	schedule();
	}
		
	set_current_state(TASK_RUNNING);
	remove_wait_queue(&info->open_wait, &wait);
	
 	if (!tty_hung_up_p(filp)) 
 		info->count++;
		

	info->blocked_open--;
	
	if (retval) 
		return(retval);

	info->flags |= ASYNC_NORMAL_ACTIVE;
	return(0);			
}


static void sunix_shutdown(struct sunix_port_hwconf *info)
{
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,11))	
	unsigned long flag;
#endif

	if(!(info->flags & ASYNC_INITIALIZED)) 
		return;

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,11))	
	spin_lock_irqsave(&info->slock, flag);
#endif

	wake_up_interruptible(&info->delta_msr_wait);
	
	if (info->tx_buf) 
	{
		free_page((unsigned long)info->tx_buf);
		info->tx_buf = NULL;
	}
	
	info->IER = 0;
	outb(0x00 , info->base + UART_IER);
	
	if (!info->tty || (info->tty->termios->c_cflag & HUPCL))
		info->MCR &= ~(UART_MCR_DTR | UART_MCR_RTS);

	outb(info->MCR, info->base + UART_MCR);		
	
	outb((UART_FCR_CLEAR_RCVR|UART_FCR_CLEAR_XMIT), info->base + UART_FCR);
	
	(void)inb(info->base + UART_RX);

	if (info->tty)
		set_bit(TTY_IO_ERROR, &info->tty->flags);

	info->flags &= ~ASYNC_INITIALIZED;

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,11))	
	spin_unlock_irqrestore(&info->slock,flag);
#endif
}


static void sunix_wait_until_sent(struct tty_struct *tty, int timeout)
{
	struct sunix_port_hwconf *info = (struct sunix_port_hwconf *)tty->driver_data;
	unsigned long orig_jiffies, char_time;
	int lsr;
	
	if (info->type == PORT_UNKNOWN)
		return;

	if (info->xmit_fifo_size == 0) 
		return; 
		
	orig_jiffies = jiffies;
	char_time = (info->timeout - HZ/50) / info->xmit_fifo_size;
	char_time = char_time / 5;

	if (char_time == 0)
		char_time = 1;
	if (timeout && timeout < char_time)
		char_time = timeout;	
	
	if (!timeout || timeout > 2*info->timeout)
		timeout = 2*info->timeout;	
	
	while (!((lsr = inb(info->base+ UART_LSR)) & UART_LSR_TEMT)) 
	{
		set_current_state(TASK_INTERRUPTIBLE);
		schedule_timeout(char_time);
		if (signal_pending(current)) 
			break;
			
		if (timeout && time_after(jiffies, orig_jiffies + timeout)) 
			break;
	}

	set_current_state(TASK_RUNNING);		
}


static void sunix_flush_chars(struct tty_struct * tty)
{
	struct sunix_port_hwconf *info = (struct sunix_port_hwconf *)tty->driver_data;


	if(info->tx_buf_cont<=0||tty->stopped||!info->tx_buf||(tty->hw_stopped)) 
		return;


		
	info->IER |= UART_IER_THRI;
	outb(info->IER, info->base + UART_IER);

}


static int sunix_open(struct tty_struct *tty, struct file *filp)
{
	struct sunix_port_hwconf *info;
	int retval, line;

 	line = tty->index;

	if ((line < 0) || (line >= SUNIX_MAX_PORTS_TOTAL)) 
		return(-ENODEV);

	info = sunix_port_table + line;	
	
	if (!info->base) 
	    	return(-ENODEV);

 	tty->driver_data = info;
  	info->tty = tty;

   	retval = sunix_startup(info);
   	if (retval)
   		return(retval);
		
	
  	retval = sunix_block_til_ready(tty, filp, info);
  	if (retval)
  		return(retval);		
		
	info->count++;

	try_module_get(THIS_MODULE);

	
	
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,5))	
	info->session = current->signal->session;
#else
	info->session = current->session;	
#endif			

 	info->pgrp = process_group(current);
	clear_bit(TTY_DONT_FLIP, &tty->flags);	 

 	tty->low_latency = 1;	
	return (0);
}

static void sunix_close(struct tty_struct *tty, struct file *filp)
{
	struct sunix_port_hwconf *info = (struct sunix_port_hwconf *)tty->driver_data;
 	unsigned long timeout;
	int line;
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,11))	
	unsigned long flag;
#endif
	
	line = tty->index;

	if ((line < 0) || (line >= SUNIX_MAX_PORTS_TOTAL)) 
		return;
				
	if (!info) 
	    	return;
		
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,11))	
	spin_lock_irqsave(&info->slock, flag);
#endif

		
	if(tty_hung_up_p(filp)) 
	{
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,11))	
		spin_unlock_irqrestore(&info->slock,flag);
#endif
 		module_put(THIS_MODULE);
		return;
	}			
		
	if ((tty->count == 1) && (info->count != 1)) 
	    	info->count = 1;
		
	if (--info->count < 0) 
	    	info->count = 0;
								
	if (info->count) 
	{
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,11))	
		spin_unlock_irqrestore(&info->slock,flag);
#endif
 	    	module_put(THIS_MODULE);
	    	return;
	}
		
	info->flags |= ASYNC_CLOSING;
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,11))	
	spin_unlock_irqrestore(&info->slock,flag);
#endif
		
	if (info->flags & ASYNC_NORMAL_ACTIVE) 
	    	info->normal_termios = *tty->termios;
		
		
	tty->closing = 1;
	
	if (info->closing_wait != ASYNC_CLOSING_WAIT_NONE) 
		tty_wait_until_sent(tty, info->closing_wait);
			
		
	info->IER &= ~UART_IER_RLSI;	
	
	if (info->flags & ASYNC_INITIALIZED) 
	{
		outb(info->IER, info->base + UART_IER);
	    	timeout = jiffies + HZ;
	
	    	while (!(inb(info->base + UART_LSR) & UART_LSR_TEMT)) 
		{
			set_current_state(TASK_INTERRUPTIBLE);
			schedule_timeout(5);
			
			if (time_after(jiffies, timeout)) 
		    		break;
	    	}
	}	

	sunix_shutdown(info);

		
	if (tty->driver->flush_buffer)
		tty->driver->flush_buffer(tty);

	if (tty->ldisc.flush_buffer)
		tty->ldisc.flush_buffer(tty);

		
				
	tty->closing 	= 0;
	info->event 	= 0;
	info->tty 	= NULL;
	
	if (info->blocked_open) 
	{
	    	if (info->close_delay) 
		{
			set_current_state(TASK_INTERRUPTIBLE);
			schedule_timeout(info->close_delay);
	    	}
	    	wake_up_interruptible(&info->open_wait);
	}

	info->flags &= ~(ASYNC_NORMAL_ACTIVE | ASYNC_CLOSING);
	
	wake_up_interruptible(&info->close_wait);
	module_put(THIS_MODULE);
}	



#if (LINUX_VERSION_CODE > KERNEL_VERSION(2,6,10))
static int sunix_write(struct tty_struct *tty, const unsigned char *buf, int count)
#else
static int sunix_write(struct tty_struct *tty, int from_user, const unsigned char *buf, int count)
#endif
{
	int c, total = 0;
	struct sunix_port_hwconf *info = (struct sunix_port_hwconf *)tty->driver_data;
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,11))	
	unsigned long flag;
#endif

	if (!tty || !info->tx_buf) 
	    	return(0);
	
	while (1) 
	{
		c = MIN(count, MIN(SERIAL_XMIT_SIZE - info->tx_buf_cont - 1, SERIAL_XMIT_SIZE - info->tx_buf_head));
	 	if (c <= 0)
			break;
				
#if (LINUX_VERSION_CODE > KERNEL_VERSION(2,6,9))				
 		memcpy(info->tx_buf + info->tx_buf_head , buf, c);
#else
	    	if (from_user) 
		{
			if(copy_from_user(info->tx_buf + info->tx_buf_head , buf, c) == c)
			{
				total = -EFAULT;	
		    		break;
			}
	    	} 
		else 
			memcpy(info->tx_buf + info->tx_buf_head , buf, c);

#endif							
			
			
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,11))	
		spin_lock_irqsave(&info->slock, flag);
#endif
	    	info->tx_buf_head   = (info->tx_buf_head + c) & (SERIAL_XMIT_SIZE - 1);
	    	info->tx_buf_cont  += c;

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,11))	
		spin_unlock_irqrestore(&info->slock,flag);
#endif

	    	buf   += c;
	    	count -= c;
	  	total += c;
	}

 	if (info->tx_buf_cont && !tty->stopped && !(info->IER & UART_IER_THRI)) 
	{
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,11))	
		spin_lock_irqsave(&info->slock, flag);
#endif
            	info->IER |= UART_IER_THRI;
             	outb(info->IER, info->base + UART_IER);
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,11))	
		spin_unlock_irqrestore(&info->slock,flag);
#endif
	}

	return total;
}


static void sunix_put_char(struct tty_struct * tty, unsigned char ch)
{
	struct sunix_port_hwconf *info = (struct sunix_port_hwconf *)tty->driver_data;
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,11))	
	unsigned long flag;
#endif

	
	if (!tty || !info->tx_buf)
	    	return;

	if (info->tx_buf_cont >= (SERIAL_XMIT_SIZE - 1)) 
	    	return;

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,11))	
	spin_lock_irqsave(&info->slock, flag);
#endif
	info->tx_buf[info->tx_buf_head++] = ch;
	info->tx_buf_head &= SERIAL_XMIT_SIZE - 1;
	info->tx_buf_cont++;
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,11))	
	spin_unlock_irqrestore(&info->slock,flag);
#endif
	
	if (!tty->stopped && !(info->IER & UART_IER_THRI)) 
	{
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,11))	
		spin_lock_irqsave(&info->slock, flag);
#endif
            	info->IER |= UART_IER_THRI;
            	outb(info->IER, info->base + UART_IER);
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,11))	
		spin_unlock_irqrestore(&info->slock,flag);
#endif
	}	
}


static int sunix_write_room(struct tty_struct * tty)
{
	struct sunix_port_hwconf *info = (struct sunix_port_hwconf *)tty->driver_data;
	int ret;

	ret = SERIAL_XMIT_SIZE - info->tx_buf_cont - 1;
	
	if (ret < 0)
		ret = 0;
		
	return(ret);
}


static int sunix_chars_in_buffer(struct tty_struct * tty)
{
	struct sunix_port_hwconf *info = (struct sunix_port_hwconf *)tty->driver_data;
	return(info->tx_buf_cont);	
}


static void sunix_flush_buffer(struct tty_struct * tty)
{
	struct sunix_port_hwconf *info = (struct sunix_port_hwconf *)tty->driver_data;
	
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,11))	
	unsigned long flag;
#endif	
	
	
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,11))
	spin_lock_irqsave(&info->slock, flag);	
#endif	
	info->tx_buf_cont = info->tx_buf_head = info->tx_buf_tail = 0;
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,11))	
	spin_unlock_irqrestore(&info->slock,flag);
#endif		
	
	outb((info->FCR | UART_FCR_CLEAR_RCVR | UART_FCR_CLEAR_XMIT), info->base + UART_FCR);
	outb(info->FCR, info->base + UART_FCR);
		
	wake_up_interruptible(&tty->write_wait);
	
	if ((tty->flags & (1 << TTY_DO_WRITE_WAKEUP)) && tty->ldisc.write_wakeup) 
		(tty->ldisc.write_wakeup)(tty);
}


static void sunix_throttle(struct tty_struct * tty)
{
	struct sunix_port_hwconf *info = (struct sunix_port_hwconf *)tty->driver_data;

        info->ldisc_stop_rx = 1;
	
	if (info->FC_flag & SUNIX_FLOWCTL_SW) 
	{
		info->x_char = STOP_CHAR(tty);
 		outb(info->x_char, info->base + UART_TX);
		info->icount.tx++;
		info->IER |= UART_IER_THRI;		
		outb(info->IER, info->base + UART_IER);
	}

	if (info->FC_flag & SUNIX_FLOWCTL_HW_AUTO) 
	{
 		info->MCR &= ~UART_MCR_AFE;
 		info->MCR &= ~UART_MCR_RTS;
		outb(info->MCR, info->base + UART_MCR);
	}	
}


static void sunix_unthrottle(struct tty_struct * tty)
{
	struct sunix_port_hwconf *info = (struct sunix_port_hwconf *)tty->driver_data;
        info->ldisc_stop_rx = 0;	

        if (info->FC_flag & SUNIX_FLOWCTL_SW) 
	{
		info->x_char = START_CHAR(tty);
 		outb(info->x_char, info->base + UART_TX);
		info->icount.tx++;
		info->IER |= UART_IER_THRI;		
		outb(info->IER, info->base + UART_IER);
	}

	if (info->FC_flag & SUNIX_FLOWCTL_HW_AUTO) 
	{
 		info->MCR |= UART_MCR_AFE;
 		info->MCR |= UART_MCR_RTS;
		outb(info->MCR, info->base + UART_MCR);
	}
}


static void sunix_stop(struct tty_struct * tty)
{
	struct sunix_port_hwconf *info = (struct sunix_port_hwconf *)tty->driver_data;
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,11))	
	unsigned long flag;
#endif

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,11))	
	spin_lock_irqsave(&info->slock, flag);
#endif
	if (info->IER & UART_IER_THRI) 
	{
	    	info->IER &= ~UART_IER_THRI;
	    	outb(info->IER, info->base + UART_IER);
	}
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,11))	
	spin_unlock_irqrestore(&info->slock,flag);
#endif

}


static void sunix_start(struct tty_struct * tty)
{
	struct sunix_port_hwconf *info = (struct sunix_port_hwconf *)tty->driver_data;
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,11))	
	unsigned long flag;
#endif

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,11))	
	spin_lock_irqsave(&info->slock, flag);
#endif
	if (info->tx_buf_cont && info->tx_buf && !(info->IER & UART_IER_THRI)) 
	{
	    	info->IER |= UART_IER_THRI;
	    	outb(info->IER, info->base + UART_IER);
	}
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,11))	
	spin_unlock_irqrestore(&info->slock,flag);
#endif

}


static void sunix_hangup(struct tty_struct * tty)
{
	struct sunix_port_hwconf *info = (struct sunix_port_hwconf *)tty->driver_data;
	
	sunix_flush_buffer(tty);
	sunix_shutdown(info);
	info->event = 0;
	info->count = 0;
	info->flags &= ~(ASYNC_NORMAL_ACTIVE);
	info->tty = NULL;
	wake_up_interruptible(&info->open_wait);
}


static void sunix_rs_break(struct tty_struct *tty, int break_state)
{

	struct sunix_port_hwconf *info = (struct sunix_port_hwconf *)tty->driver_data;
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,11))	
	unsigned long flag;
#endif


#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,11))	
	spin_lock_irqsave(&info->slock, flag);
#endif
	if (break_state == -1)
		outb(inb(info->base + UART_LCR) |  UART_LCR_SBC, info->base + UART_LCR);
	else
		outb(inb(info->base + UART_LCR) & ~UART_LCR_SBC, info->base + UART_LCR);

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,11))	
	spin_unlock_irqrestore(&info->slock,flag);
#endif

}


#if (LINUX_VERSION_CODE > KERNEL_VERSION(2,6,10))
static void sunix_modem_status_intr(struct sunix_port_hwconf *info, unsigned char msr)
#else
static inline void sunix_modem_status_intr(struct sunix_port_hwconf *info, unsigned char msr)
#endif
{
	if ( msr & UART_MSR_TERI )
	    info->icount.rng++;
	if ( msr & UART_MSR_DDSR )
	    info->icount.dsr++;
	if ( msr & UART_MSR_DDCD )
	    info->icount.dcd++;
	if ( msr & UART_MSR_DCTS )
	    info->icount.cts++;	
	
	wake_up_interruptible(&info->delta_msr_wait);
	
	if ((info->flags & ASYNC_CHECK_CD) && (msr & UART_MSR_DDCD)) 
	{
	    	if ( msr & UART_MSR_DCD )
			wake_up_interruptible(&info->open_wait);
		else if ( !((info->flags & ASYNC_CALLOUT_ACTIVE) && (info->flags & ASYNC_CALLOUT_NOHUP)) )
	        	set_bit(SUNIX_EVENT_HANGUP,&info->event);
			
            	schedule_work(&info->tqueue);
	}

	if (info->flags & SUNIX_FLOWCTL_HW_AUTO) 
	{
		if (info->tty->hw_stopped) 
		{
			if (msr & UART_MSR_CTS)
			{
		    		info->tty->hw_stopped = 0;

		    		if ((info->type != PORT_16650) && (!info->SunixChipFlag))
				{
					info->IER |= UART_IER_THRI;
					outb(info->IER, info->base + UART_IER);
		    		}
	       			set_bit(SUNIX_EVENT_TXLOW,&info->event);
	       			schedule_work(&info->tqueue);
        		}
	    	}
		else 
		{
			if (!(msr & UART_MSR_CTS))
			{
			    	info->tty->hw_stopped = 1;
		    		if ((info->type != PORT_16650) && (!info->SunixChipFlag)) 
				{
					info->IER &= ~UART_IER_THRI;
					outb(info->IER, info->base + UART_IER);
		    		}
			}
	    	}
	}
}



#if (LINUX_VERSION_CODE > KERNEL_VERSION(2,6,10))
static void sunix_data_transmit(struct sunix_port_hwconf *info)
#else
static inline void sunix_data_transmit(struct sunix_port_hwconf *info)
#endif
{
	int count, cnt, tail_tmp = 0;
	unsigned char tmp;
	struct tty_struct *tty;
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,11))	
	unsigned long flag;
#endif
	
	tty = info->tty;		
	
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,11))	
	spin_lock_irqsave(&info->slock, flag);
#endif



	if ((info->tx_buf == 0))
	{
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,11))	
		spin_unlock_irqrestore(&info->slock,flag);
#endif
	    	return;
	}	

	if ((info->tx_buf_cont<=0) || (info->tty->stopped) || (tty->hw_stopped)) 
	{
		info->IER &= ~UART_IER_THRI;
		outb(info->IER, info->base + UART_IER);
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,11))	
		spin_unlock_irqrestore(&info->slock,flag);
#endif
		return;
	}
 
	cnt     = info->tx_buf_cont;
	count   = info->xmit_fifo_size; 	
	
	do {
		tail_tmp = info->tx_buf_tail;
		tmp = info->tx_buf[info->tx_buf_tail++];
		outb(tmp, info->base + UART_TX);
	    	info->tx_buf_tail = info->tx_buf_tail & (SERIAL_XMIT_SIZE - 1);
			
	    	if ((--info->tx_buf_cont <= 0))
			break;
			
	} while (--count > 0);

	info->icount.tx += (cnt - info->tx_buf_cont);
		
  	if (info->tx_buf_cont < WAKEUP_CHARS) 
	{
  		set_bit(SUNIX_EVENT_TXLOW, &info->event);
 		schedule_work(&info->tqueue);
 	}	

	if (info->tx_buf_cont <= 0) 
	{
		info->IER &= ~UART_IER_THRI;
		outb(info->IER, info->base + UART_IER);
	}
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,11))	
	spin_unlock_irqrestore(&info->slock,flag);
#endif
}



#if (LINUX_VERSION_CODE > KERNEL_VERSION(2,6,10))
static void sunix_data_receive(struct sunix_port_hwconf *info, int *status, unsigned char iir)
#else
static inline void sunix_data_receive(struct sunix_port_hwconf *info, int *status, unsigned char iir)
#endif
{
	struct tty_struct *tty = info->tty;
	unsigned char ch = 0x00, lsr = 0x00;
	int count, ignored = 0, recv_room = 0;
 	unsigned char 			*cp;
 	char 				*fp;
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,11))	
	unsigned long flag;
#endif
	
	recv_room = tty->ldisc.receive_room(tty);

     	if ((recv_room == 0) && (!info->ldisc_stop_rx)) 
	{
    		sunix_throttle(tty);
		return;
    	}

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,11))	
	spin_lock_irqsave(&info->slock, flag);
#endif

  	cp = tty->flip.char_buf;
  	fp = tty->flip.flag_buf;
        count = 0;	

	if (iir == UART_IIR_RLSI) 
		goto RLSI_START;
		
	if ((iir == UART_IIR_RDI) || (iir == UART_IIR_CTO)) 
	{
		do{
			ch = inb(info->base + UART_RX);
			info->icount.rx++;				
			
			if (info->FC_flag & SUNIX_FLOWCTL_SW) 
			{
				if(ch == STOP_CHAR(info->tty))
				{
					info->tty->stopped = 1;
  					info->IER |= UART_IER_THRI;		
  					outb(info->IER, info->base + UART_IER);
					goto END_SWF;
				}
				if(ch == START_CHAR(info->tty))
				{
					info->tty->stopped = 0;
   					info->IER |= UART_IER_THRI;		
   					outb(info->IER, info->base + UART_IER);
					goto END_SWF;
				}	
			}		

			*cp++ = ch;
 			*fp++ = 0;
			count++;
END_SWF:
			if ((iir == UART_IIR_RDI) && (count < 15))
				lsr = 0x01;
			else
				lsr = inb(info->base + UART_LSR);
				
		} while (lsr & UART_LSR_DR);	
	}			
	goto RECE_END;
	
	
RLSI_START:

	do{
	    	ch = inb(info->base + UART_RX);
		info->icount.rx++;
		if (*status & (UART_LSR_BI | UART_LSR_PE | UART_LSR_FE | UART_LSR_OE)) 
		{
			
			if (*status & info->ignore_status_mask) 
			{
				if (++ignored > 100)
					break;
				goto IGNORE_CHAR;
			}	

			*status &= info->read_status_mask;

			if (*status & UART_LSR_OE)
			{ 
				*fp++ = TTY_OVERRUN;
				info->icount.overrun++;
				outb((info->FCR | 0x23), info->base + UART_FCR);
			}
			else if (*status & UART_LSR_BI) 
			{
				info->icount.brk++;
				*fp++ = TTY_BREAK;
				if (info->flags & ASYNC_SAK)
					do_SAK(tty);
			}
			else if (*status & UART_LSR_PE) 
			{
				*fp++ = TTY_PARITY;
				info->icount.parity++;
			}
			else if(*status & UART_LSR_FE) 
			{
				*fp++ = TTY_FRAME;
				info->icount.frame++;
			}
			else	
				*fp++ = 0;
		}
		*cp++ = ch;
		count++;
IGNORE_CHAR:
		*status = inb(info->base+UART_LSR);
		
	} while (*status & UART_LSR_DR);		

RECE_END:	
	if(count > 0)	
		tty->ldisc.receive_buf(tty, tty->flip.char_buf, tty->flip.flag_buf, count);
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,11))	
	spin_unlock_irqrestore(&info->slock,flag);
#endif
}


static irqreturn_t sunix_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
	unsigned char msr;
	int status, iir, i, max, irqbits, bits, pass_counter = 0, handled = IRQ_NONE;
	struct sunix_port_hwconf *info, *first_port;


	first_port = NULL;		
        for(i = 0; i < SUNIX_MAX_BOARDS; i++) 
	{
        	if(dev_id == &(sunix_port_table[i * SUNIX_MAX_PORTS_PER_BOARD])) 
		{
                	first_port = dev_id;
			break;
            	}
        }
			
        if(i == SUNIX_MAX_BOARDS)
            	goto INTR_END;
		
        if( first_port == 0) 
            	goto INTR_END;
	
        max = sunix_numports[sunix_pci_board[i].board_type-1];
	
	while (1) {
	    		irqbits = inb(first_port->vector) & first_port->vector_mask;
			
	    		if (irqbits == 0x00)
				break;  

	    		handled = IRQ_HANDLED;
	    		for (i = 0, bits = 1; i < max; i++, bits <<= 1) 
			{
				
				if (!(bits & irqbits))
					continue;
					
				info = first_port + i; 

				iir = inb(info->base + UART_IIR) & 0x0f;
				if (iir & UART_IIR_NO_INT)
					continue;

				status = inb(info->base + UART_LSR);

				msr = inb(info->base + UART_MSR);
				if (msr & UART_MSR_ANY_DELTA)  
 					sunix_modem_status_intr(info, msr);	
											
										
				if (iir == UART_IIR_RLSI || iir == UART_IIR_CTO || iir == UART_IIR_RDI)
					sunix_data_receive(info, &status, iir);					
				
				if ((iir == UART_IIR_THRI) && (status & UART_LSR_THRE))  
					sunix_data_transmit(info);
				}
	    		if (pass_counter++ > 128) 
				break;	
		}
INTR_END:
	return handled;
}

static void sunix_send_break(struct sunix_port_hwconf * info, int duration)
{
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,11))	
	unsigned long flag;
#endif

	if (!info->base)
		return;
	set_current_state(TASK_INTERRUPTIBLE);

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,11))	
	spin_lock_irqsave(&info->slock, flag);
#endif

	outb(inb(info->base + UART_LCR) | UART_LCR_SBC, info->base + UART_LCR);
	schedule_timeout(duration);
	outb(inb(info->base + UART_LCR) & ~UART_LCR_SBC, info->base + UART_LCR);

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,11))	
	spin_unlock_irqrestore(&info->slock,flag);
#endif

}


static int sunix_tiocmget(struct tty_struct *tty, struct file *file)
{
	struct sunix_port_hwconf *info = (struct sunix_port_hwconf *)tty->driver_data;
	unsigned char mcr, msr;
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,11))	
	unsigned long flag;
#endif


	if (tty->index >= SUNIX_MAX_PORTS_TOTAL)
		return (-ENOIOCTLCMD);
		
	if (tty->flags & (1 << TTY_IO_ERROR))
		return (-EIO);

	mcr = info->MCR;

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,11))	
	spin_lock_irqsave(&info->slock, flag);
#endif	

	msr = inb(info->base + UART_MSR);
	if ( msr & UART_MSR_ANY_DELTA )
		sunix_modem_status_intr(info, msr);

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,11))	
	spin_unlock_irqrestore(&info->slock,flag);
#endif	

	return 	((mcr & UART_MCR_RTS) ? TIOCM_RTS : 0) |
	    	((mcr & UART_MCR_DTR) ? TIOCM_DTR : 0) |
	    	((msr & UART_MSR_DCD) ? TIOCM_CAR : 0) |
	    	((msr & UART_MSR_RI ) ? TIOCM_RNG : 0) |
	    	((msr & UART_MSR_DSR) ? TIOCM_DSR : 0) |
	    	((msr & UART_MSR_CTS) ? TIOCM_CTS : 0);
}


static int sunix_tiocmset(struct tty_struct *tty, struct file *file,unsigned int set, unsigned int clear)
{
	struct sunix_port_hwconf *info = (struct sunix_port_hwconf *)tty->driver_data;
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,11))	
	unsigned long flag;
#endif


	if (tty->index >= SUNIX_MAX_PORTS_TOTAL)
		return (-ENOIOCTLCMD);
		
	if (tty->flags & (1 << TTY_IO_ERROR))
		return (-EIO);


#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,11))	
	spin_lock_irqsave(&info->slock, flag);
#endif
	
	if (set & TIOCM_RTS) 
		info->MCR |= UART_MCR_RTS;
		
	if (set & TIOCM_DTR)
		info->MCR |= UART_MCR_DTR;

	if (clear & TIOCM_RTS)
		info->MCR &= ~UART_MCR_RTS;
		
	if (clear & TIOCM_DTR)
		info->MCR &= ~UART_MCR_DTR;
	
	outb(info->MCR, info->base + UART_MCR);
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,11))	
	spin_unlock_irqrestore(&info->slock,flag);
#endif
	return 0;	
}


static int sunix_get_uart_info(struct sunix_port_hwconf *info, struct serial_struct *retinfo)
{
	struct serial_struct tmp;

	if (!retinfo)
	    return(-EFAULT);
	    
	memset(&tmp, 0, sizeof(tmp));
	
	tmp.type 		= info->type;
	tmp.line 		= info->port;
	tmp.port 		= info->base;
	tmp.irq  		= info->irq;
	tmp.flags 		= info->flags;
	tmp.baud_base 		= info->baud_base;
	tmp.close_delay 	= info->close_delay;
	tmp.closing_wait 	= info->closing_wait;
	tmp.custom_divisor 	= info->custom_divisor;
	tmp.hub6 		= 0;
	
	if (copy_to_user(retinfo, &tmp, sizeof(*retinfo)))
	    return -EFAULT;
	return 0;
}


static int sunix_set_uart_info(struct sunix_port_hwconf *info, struct serial_struct *new_info)
{
	struct serial_struct new_serial;
	unsigned int flags;
	int retval = 0;

	if (!new_info || !info->base)
		return(-EFAULT);
	if(copy_from_user(&new_serial, new_info, sizeof(new_serial)))
	    	return -EFAULT;

	if (	(new_serial.irq != info->irq) || 
		(new_serial.port != info->base) || 
		(new_serial.custom_divisor != info->custom_divisor) || 
		(new_serial.baud_base != info->baud_base))
			return(-EPERM);

	flags = info->flags & ASYNC_SPD_MASK;

       	if (!capable(CAP_SYS_ADMIN)) 
	{
		if (	(new_serial.baud_base != info->baud_base) || 
			(new_serial.close_delay != info->close_delay) || 
			((new_serial.flags & ~ASYNC_USR_MASK) != (info->flags & ~ASYNC_USR_MASK)))
				return(-EPERM);
	    	info->flags = ((info->flags & ~ASYNC_USR_MASK) | (new_serial.flags & ASYNC_USR_MASK));
	}
	else 
	{
	    	info->flags = ((info->flags & ~ASYNC_FLAGS) | (new_serial.flags & ASYNC_FLAGS));
	    	info->close_delay = new_serial.close_delay * HZ/100;
	    	info->closing_wait = new_serial.closing_wait * HZ/100;
	    	info->tty->low_latency = (info->flags & ASYNC_LOW_LATENCY) ? 1 : 0;
	    	info->tty->low_latency = 0;
	    	info->type = new_serial.type;

	}

	if ( info->SunixChipFlag == SUNIX_1889_HWID) 
	{
		info->xmit_fifo_size	= 32;		
		info->rx_trigger	= 16;
	}
	else if ( info->SunixChipFlag == SUNIX_1699_HWID) 
	{
		info->xmit_fifo_size	= 64;		
		info->rx_trigger	= 32;
	}		
	else
	{
		info->xmit_fifo_size	= 1;		
		info->rx_trigger	= 1;
	}	


	if (info->flags & ASYNC_INITIALIZED) 
	{
	    	if (flags != (info->flags & ASYNC_SPD_MASK))
			termios_setup(info,0);
	}
	else
	    	retval = sunix_startup(info);
		
		
	return(retval);
}


static int sunix_get_lsr_info(struct sunix_port_hwconf *info, unsigned int *value)
{
	unsigned char lsr;
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,11))	
	unsigned long flag;
	spin_lock_irqsave(&info->slock, flag);
#endif

	lsr = inb(info->base + UART_LSR);

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,11))	
	spin_unlock_irqrestore(&info->slock,flag);
#endif
	if (copy_to_user(value, &lsr, sizeof(int)))
		return -EFAULT;
	return 0;
}


static void sunix_set_termios(struct tty_struct *tty, struct termios *old_termios)
{
	struct sunix_port_hwconf *info = (struct sunix_port_hwconf *)tty->driver_data;
	
	if((tty->termios->c_cflag != old_termios->c_cflag)|| (RELEVANT_IFLAG(tty->termios->c_iflag)!=RELEVANT_IFLAG(old_termios->c_iflag)))
		termios_setup(info, old_termios);
		

	if ((old_termios->c_cflag & CRTSCTS) && !(tty->termios->c_cflag & CRTSCTS)) 
	{
 	   	tty->hw_stopped = 0;
		sunix_start(tty);
	}

	if ((old_termios->c_iflag & IXON) && !(tty->termios->c_iflag & IXON)) 
	{
 		tty->stopped = 0;
     		sunix_start(tty);
	}
}


static int sunix_ioctl(struct tty_struct * tty, struct file * file,unsigned int cmd, unsigned long arg)
{
 	struct sunix_port_hwconf *info = (struct sunix_port_hwconf *)tty->driver_data;
 	int retval,tmp,ret1;
 	struct async_icount cprev, cnow;
 	struct serial_icounter_struct icount;
	struct port_info dump;
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,11))	
	unsigned long flag;
#endif



	if ((cmd != TIOCGSERIAL) && (cmd != TIOCMIWAIT) && (cmd != TIOCGICOUNT)) 
	{
		if (tty->flags & TTY_IO_ERROR) 
			return(-EIO);
			
	}
	
	switch (cmd) 
	{

		case TCSBRK: 
	    		retval = tty_check_change(tty);
	   		if (retval)
				return(retval);
				
	    		tty_wait_until_sent(tty, 0);

			if (signal_pending(current))
				return -EINTR;				
		
	    		if (!arg) 
			{
				sunix_send_break(info, HZ/4);
				if (signal_pending(current))		
					return -EINTR;
			}
			
			return 0;
		
		case TCSBRKP:	
	   		retval = tty_check_change(tty);
	   		if (retval)
				return(retval);
			
	   		tty_wait_until_sent(tty, 0);
		
			if (signal_pending(current))
				return -EINTR;	
						
	    		sunix_send_break(info, arg ? arg*(HZ/10) : HZ/4);			
			if (signal_pending(current))
				return -EINTR;
							
			return 0;
		
		case TIOCGSOFTCAR:  
			tmp = C_CLOCAL(tty) ? 1 : 0;
			ret1 = copy_to_user((void *)arg, &tmp, sizeof(int));
			if (ret1)
				return -EFAULT;		
				
			return 0;
		
		case TIOCSSOFTCAR: 
			if (copy_from_user(&tmp, (void *)arg, sizeof(int)))
				return -EFAULT;

			tty->termios->c_cflag =	((tty->termios->c_cflag & ~CLOCAL) | (tmp ? CLOCAL : 0));				
			return 0;
		
		case TIOCGSERIAL:  
			return sunix_get_uart_info(info,(struct serial_struct *) arg);
		
		case TIOCSSERIAL: 
			return sunix_set_uart_info(info,(struct serial_struct *) arg);

		case TIOCSERGETLSR: 
			return sunix_get_lsr_info(info, (unsigned int *) arg);


		case TIOCMIWAIT:
		{  
 	    		DECLARE_WAITQUEUE(wait, current);
			int ret;

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,11))	
			spin_lock_irqsave(&info->slock, flag);
#endif
	    		cprev = info->icount;  
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,11))	
			spin_unlock_irqrestore(&info->slock,flag);
#endif
	
			add_wait_queue(&info->delta_msr_wait, &wait);
	      		while (1) 
			{
			
				cnow = info->icount;	
			
				set_current_state(TASK_INTERRUPTIBLE);
				if (	 ((arg & TIOCM_RNG) && (cnow.rng != cprev.rng)) ||
		    			 ((arg & TIOCM_DSR) && (cnow.dsr != cprev.dsr)) ||
		     			 ((arg & TIOCM_CD)  && (cnow.dcd != cprev.dcd)) ||
		     			 ((arg & TIOCM_CTS) && (cnow.cts != cprev.cts)) ) 
				{
		    			ret = 0;
		    			break;
				}
//#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,13))				
				schedule();
//#endif			

				if (signal_pending(current))
				{
		   			ret = -ERESTARTSYS;
		    			break;
				}				
				cprev = cnow;			
			}
			current->state = TASK_RUNNING;
	    		remove_wait_queue(&info->delta_msr_wait, &wait);
   	    		break;		
		}
		
		
		case TIOCGICOUNT: 
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,11))	
			spin_lock_irqsave(&info->slock, flag);
#endif
			cnow = info->icount;
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,11))	
			spin_unlock_irqrestore(&info->slock,flag);
#endif
			icount.cts         = cnow.cts;
			icount.dsr         = cnow.dsr;
			icount.rng         = cnow.rng;
			icount.dcd         = cnow.dcd;
			icount.rx          = cnow.rx;
			icount.tx          = cnow.tx;
			icount.frame       = cnow.frame;
			icount.overrun     = cnow.overrun;
			icount.parity 	   = cnow.parity;
			icount.brk         = cnow.brk;
			icount.buf_overrun = cnow.buf_overrun;
			
			if (copy_to_user((void *)arg, &icount, sizeof(icount)))
				return -EFAULT;
				
			return 0;					
		
		case SUNIX_DUMP_PORT_INFO: 
			dump.board_name_info = info->board_name;
			dump.bus_number_info = info->bus_number;
			dump.dev_number_info = info->dev_number;
			dump.port_info	     = info->port;
			dump.base_info	     = info->base;
			dump.irq_info 	     = info->irq;

			if (copy_to_user((void *)arg, &dump, sizeof(dump)))
				return -EFAULT;
			return 0;	
								
		default:
	    		return(-ENOIOCTLCMD);
	}
	return 0;
}


static void sunix_do_softint(void *private_)
{
	struct sunix_port_hwconf *info = (struct sunix_port_hwconf *)private_;	
	struct tty_struct *tty;
	
	tty = info->tty;
	
	if (tty) 
	{
		if (test_and_clear_bit(SUNIX_EVENT_TXLOW, &info->event)) 
		{

	        	if ((tty->flags &  (1 << TTY_DO_WRITE_WAKEUP)) && tty->ldisc.write_wakeup) 
				(tty->ldisc.write_wakeup)(tty);

	        	wake_up_interruptible(&tty->write_wait);
	    	}
			
	    	if (test_and_clear_bit(SUNIX_EVENT_HANGUP, &info->event))
			tty_hangup(tty);
	}
}


static int sunix_get_PCI_conf(int board_type, struct sunix_board_hwconf *hwconf)
{
	int i;
	unsigned int ioaddress[SUNIX_MAX_PORTS_PER_BOARD];
	struct pci_dev *pdev = hwconf->pdev;

	hwconf->board_type = board_type;
	hwconf->ports      = sunix_numports[board_type-1];

	for(i = 0; i < hwconf->ports; i++) 
	{
		if(i == 0) 
		{
			ioaddress[i] = pci_resource_start(pdev, i);
			hwconf->SunixChipFlag[i] = SUNIX_1889_HWID;
		}
		else if(i == 1) 
		{
			ioaddress[i] = ioaddress[i-1]+8;
			hwconf->SunixChipFlag[i] = SUNIX_1889_HWID;			
		}
		else if(i == 2) 
		{
			ioaddress[i] = pci_resource_start(pdev, i-1);
			hwconf->SunixChipFlag[i] = SUNIX_1699_HWID;			
		}
		else if(i == 3) 
		{
			ioaddress[i] = ioaddress[i-1]+8;
			hwconf->SunixChipFlag[i] = SUNIX_1699_HWID;			
		}
		else if((i == 4)||(i == 5)||(i == 6)||(i == 7)) 
		{
			ioaddress[i] = pci_resource_start(pdev, i-2);
			hwconf->SunixChipFlag[i] = SUNIX_1699_HWID;			
		}	
	}

	
	for(i = 0; i <  hwconf->ports; i++) 
		request_region(ioaddress[i],8, "sunix");

	hwconf->vector_mask = 0;
		
		
	for (i = 0; i < hwconf->ports; i++) 
	{
		hwconf->vector_mask  |= (1 << i);
		hwconf->ioaddr[i] = ioaddress[i];
		hwconf->baud_base[i] = 921600;	
			
	}	

	hwconf->vector 		= ioaddress[0] + 0x1c;
	hwconf->irq 		= hwconf->pdev->irq;	
	hwconf->uart_type 	= PORT_16650;

	return(0);
}


static int sunix_board_init(int board,struct sunix_board_hwconf *hwconf)
{
	struct sunix_port_hwconf *info;
        int retval, i, n;

	n = (board * SUNIX_MAX_PORTS_PER_BOARD);
	info = &sunix_port_table[n];


	for (i = 0; i < hwconf->ports; i++, n++, info++) 
	{
		info->board_name 	= hwconf->board_name;
		info->board_type	= hwconf->board_type;
		info->FC_flag           = 0;
		info->bus_number	= hwconf->bus_number;
		info->dev_number	= hwconf->dev_number;
		info->count		= 0;
		info->port 		= n;		
		info->base 		= hwconf->ioaddr[i];
		info->irq 		= hwconf->irq;
		info->vector 		= hwconf->vector;
		info->vector_mask 	= hwconf->vector_mask;
		info->ldisc_stop_rx	= 0;
		info->SunixChipFlag 	= hwconf->SunixChipFlag[i];
		info->flags 		= ASYNC_SHARE_IRQ;		
		info->type 		= hwconf->uart_type;
		info->baud_base 	= hwconf->baud_base[i];	
		info->custom_divisor 	= hwconf->baud_base[i] * 16;		
		info->close_delay 	= 5*HZ/10;
		info->closing_wait 	= 30*HZ;									
		info->err_shadow	= 0;	
		info->normal_termios	= sunix_tty_driver->init_termios;		

   		INIT_WORK(&info->tqueue, sunix_do_softint, info);
		init_waitqueue_head(&info->open_wait);
		init_waitqueue_head(&info->close_wait);
		init_waitqueue_head(&info->delta_msr_wait);
		spin_lock_init(&info->slock);	
		
		if (info->SunixChipFlag == SUNIX_1889_HWID) 
		{
			info->xmit_fifo_size	= 32;		
			info->rx_trigger	= 16;
		}
		else if (info->SunixChipFlag == SUNIX_1699_HWID) 
		{
			info->xmit_fifo_size	= 32;		
			info->rx_trigger	= 16;
		}		
		else
		{
			info->xmit_fifo_size	= 1;		
			info->rx_trigger	= 1;
		}			
	}		
	
	for(i = 0; i < hwconf->ports; i++)
		outb(inb(hwconf->ioaddr[i] + UART_IER) & 0xf0, hwconf->ioaddr[i]+UART_IER);

	n = (board * SUNIX_MAX_PORTS_PER_BOARD);
	info = &sunix_port_table[n];	

	retval = request_irq(hwconf->irq, sunix_interrupt, SA_SHIRQ ,"sunix", info );
	if (retval) 
	{
	 	printk("  Board %d: %d series", board, sunix_board_name[hwconf->board_type-1]);
		printk("  Request irq fail,IRQ (%d) may be conflit with another device.\n",hwconf->irq);
		return(retval);
	}
        return 0;
}


static struct tty_operations sunix_tty_ops = 
{
	.open			= sunix_open,
	.close			= sunix_close,
	.write			= sunix_write,
 	.put_char		= sunix_put_char,
 	.flush_chars		= sunix_flush_chars,
 	.write_room		= sunix_write_room,
	.chars_in_buffer	= sunix_chars_in_buffer,
	.flush_buffer		= sunix_flush_buffer,
	.ioctl			= sunix_ioctl,
	.throttle 		= sunix_throttle,
	.unthrottle		= sunix_unthrottle,
	.set_termios 		= sunix_set_termios,
	.stop 			= sunix_stop,
	.start			= sunix_start,
	.hangup			= sunix_hangup,
	.tiocmget		= sunix_tiocmget,
	.tiocmset		= sunix_tiocmset,	
};

static int __init sunix_init(void)
{
	int retval, i, index, b=0, m=0, n=0;
	unsigned short int subdeviceID;
	struct pci_dev *pdev=NULL, *pdev_array[4] = {0,0,0,0};
	struct sunix_board_hwconf hwconf;

	printk("Loading SUNIX PCI I/O Driver Module\n");	
	sunix_tty_driver = alloc_tty_driver(SUNIX_MAX_PORTS_TOTAL);

	if (!sunix_tty_driver)
		return -ENOMEM;

	spin_lock_init(&sunix_io_lock);				
		
	for(i = 0; i < SUNIX_MAX_BOARDS; i++)
		sunix_pci_board[i].board_type = -1;
		

	memset(sunix_tty_driver,0,sizeof(struct tty_driver));	
	
	sunix_tty_driver->magic 		= TTY_DRIVER_MAGIC;	
	sunix_tty_driver->name 			= "ttySNX";
	sunix_tty_driver->major 		= SUNIX_TTY_MAJOR,
	sunix_tty_driver->minor_start 		= 0;
	sunix_tty_driver->num 			= (SUNIX_MAX_PORTS_TOTAL + 1);
	sunix_tty_driver->type 			= TTY_DRIVER_TYPE_SERIAL,
	sunix_tty_driver->subtype 		= SERIAL_TYPE_NORMAL,
	sunix_tty_driver->flags 		= TTY_DRIVER_REAL_RAW,
	sunix_tty_driver->init_termios 		= tty_std_termios;
	sunix_tty_driver->init_termios.c_cflag 	= B9600 | CS8 | CREAD | HUPCL | CLOCAL;
  	sunix_tty_driver->init_termios.c_iflag 	= 0;
	tty_set_operations(sunix_tty_driver, &sunix_tty_ops);
	
	sunix_tty_driver->ttys			= sunix_tty;
	sunix_tty_driver->termios		= sunix_termios;
	sunix_tty_driver->termios_locked	= sunix_termios_locked;

	sunix_tty_driver->break_ctl 		= sunix_rs_break;	
	sunix_tty_driver->wait_until_sent 	= sunix_wait_until_sent;	

	memset(sunix_port_table, 0, SUNIX_MAX_PORTS_TOTAL * sizeof(struct sunix_port_hwconf));
	memset(&hwconf, 0, sizeof(struct sunix_board_hwconf));	


	n = (sizeof(sunix_pcibrds) / sizeof(sunix_pcibrds[0])) - 1;
	index = 0;
	b     = 0;
	while (b < n) 
	{
		pdev = pci_find_device(PCI_VENDOR_ID_SUNIX,PCI_DEVICE_ID_SUNIX, pdev);
			
		if (pdev == NULL) 
		{
			b++;
			continue;
		}
 			
		if(m >= SUNIX_MAX_BOARDS) 
		{
 			printk("SUNIX PCI I/O Driver Module Support Four Boards In Maximum\n");	
 			break;
 		}
		else 
		{
			if( (b > 0) && ( (pdev == pdev_array[0]) ||
					 (pdev == pdev_array[1]) || 
					 (pdev == pdev_array[2]) || 
					 (pdev == pdev_array[3])) )
				continue;

			pci_read_config_word(pdev,0x2e,&subdeviceID);
			if(subdeviceID != sunix_pcibrds[b].subdevice)
				continue;
	
			if (pci_enable_device(pdev))
			{ 
				printk("SUNIX PCI Enable Fail !\n");
				continue;
			}
			pdev_array[m]     = pdev;
			hwconf.pdev       = pdev;
			hwconf.bus_number = pdev->bus->number;
 			hwconf.dev_number = PCI_SLOT(pdev->devfn);
				
			hwconf.board_name = sunix_board_name[(sunix_pcibrds[b].driver_data)-1];

			printk("Found SUNIX %d series Board , BUSNumber : %d  DEVNumber : %d\n",
			sunix_board_name[(sunix_pcibrds[b].driver_data)-1],
			hwconf.bus_number,
			hwconf.dev_number);

			retval = sunix_get_PCI_conf((int)sunix_pcibrds[b].driver_data,&hwconf);
			sunix_pci_board[m] = hwconf;
	    		sunix_board_init(m,&hwconf);	
			m++;	

		}
	}	


	retval = 0;
        if (!(retval=tty_register_driver(sunix_tty_driver))) 
	{
		printk("Loading SUNIX PCI I/O Driver Module Done\n");
		return 0;
	}	
        else
		printk("Couldn't Load SUNIX PCI I/O Driver Module\n");	
	
       if(retval)
       {
		for(i = 0; i < SUNIX_MAX_BOARDS; i++)
		{
	        	if(sunix_pci_board[i].board_type == -1)
	           		 continue;
                	else
		    		free_irq(sunix_pci_board[i].irq, &sunix_port_table[i*SUNIX_MAX_PORTS_PER_BOARD]);
            	}
		
           	return -1;
       	}	
	
	return 0;
}


static void __exit sunix_exit(void)
{
	int i, j, error=0;

	if ((error |= tty_unregister_driver(sunix_tty_driver)))
		printk("Couldn't Unload SUNIX PCI I/O Driver Module\n");

	for(i = 0; i < SUNIX_MAX_BOARDS; i++)
	{
		struct pci_dev *pdev;

		if(sunix_pci_board[i].board_type == -1)
			continue;
		else
		{
			pdev = sunix_pci_board[i].pdev;

			free_irq(sunix_pci_board[i].irq, &sunix_port_table[i*SUNIX_MAX_PORTS_PER_BOARD]);
			if(pdev != NULL)
			{ 
				for(j = 0; j < sunix_pci_board[i].ports; j++) 
					release_region(sunix_pci_board[i].ioaddr[j], 8);
			}
		}
	}
	printk("Unload SUNIX PCI I/O Driver Module\n");
}

module_init(sunix_init);
module_exit(sunix_exit);
